from openpnm.core import Base
from openpnm.utils import logging, Docorator
import numpy as np
logger = logging.getLogger(__name__)
docstr = Docorator()


@docstr.get_sectionsf('GenericAlgorithm', sections=['Parameters'])
@docstr.dedent
class GenericAlgorithm(Base):
    r"""
    Generic class to define the foundation of Algorithms.

    Parameters
    ----------
    network : (OpenPNM Network object)
        The network object to which this algorithm will apply.
    name : (string, optional)
        Name of the algorithm
    project : (OpenPNM Project object, optional)
        Either a Network or a Project must be supplied

    Notes
    -----
    This class defines the following methods, which all raise a
    ``NotImplementedError`` and must be defined by the various subclasses

    +---------------------+---------------------------------------------------+
    | Methods             | Description                                       |
    +=====================+===================================================+
    | ``results``         | Generates an array or arrays of data produced by  |
    |                     | the algorithm to be returned to the Phase         |
    +---------------------+---------------------------------------------------+
    | ``setup``           | Collects values to be placed in ``settings``. The |
    |                     | main benefit is defining default values and       |
    |                     | providing documentation on each settings          |
    +---------------------+---------------------------------------------------+
    | ``reset``           | Removes generated data, specified values, and     |
    |                     | any other information lingering on an Algorithm   |
    +---------------------+---------------------------------------------------+

    """

    def __init__(self, network=None, project=None, settings={}, **kwargs):
        self.settings.setdefault('prefix', 'alg')
        self.settings.update(settings)
        if network is not None:
            project = network.project
        super().__init__(project=project, **kwargs)

        # Deal with network or project arguments
        if network is not None:
            if project is not None:
                assert network is project.network
            else:
                project = network.project
        if project:
            self['pore.all'] = np.ones((project.network.Np, ), dtype=bool)
            self['throat.all'] = np.ones((project.network.Nt, ), dtype=bool)

    def results(self):
        r"""
        """
        raise NotImplementedError("This method must be subclassed")

    def reset(self):
        r"""
        """
        raise NotImplementedError("This method must be subclassed")

    def setup(self):
        r"""
        """
        raise NotImplementedError("This method must be subclassed")



from openpnm.phases import mixtures
import openpnm.models as mods


class Proton(mixtures.GenericSpecies):
    r"""
    Creates Phase object with preset models and values for H^+ ions

    Parameters
    ----------
    network : OpenPNM Network object
        The network to which this phase object will be attached.

    project : OpenPNM Project object, optional
        The Project with which this phase should be associted.  If a
        ``network`` is given then this is ignored and the Network's project is
        used.  If a ``network`` is not given then this is mandatory.

    name : string, optional
        The name of the phase.  This is useful to keep track of the objects
        throughout the simulation.  The name must be unique to the project.  If
        no name is given, one is generated.

    Examples
    --------
    >>> import openpnm as op
    >>> import openpnm.phases.mixtures as mixtures
    >>> pn = op.network.Cubic(shape=[5, 5, 5])
    >>> H = mixtures.species.ions.Proton(network=pn)

    """
    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self['pore.molecular_weight'] = 0.0291
        self['pore.diffusivity'] = 0.1



import pickle as pk
from openpnm.utils import logging
from openpnm.io import GenericIO
from openpnm.network import GenericNetwork
from openpnm.geometry import Imported
logger = logging.getLogger(__name__)


class PoreSpy(GenericIO):
    r"""
    """

    @classmethod
    def load(cls, filename, project=None, settings={}):
        r"""
        Load a network extracted using the PoreSpy package

        Parameters
        ----------
        filename : str or dict
            Can either be a filename point to a pickled dictionary, or an
            actual dictionary.  The second option lets users avoid the
            step of saving the dictionary to a file
        project : OpenPNM Project object
            If given, the loaded network and geometry will be added to this
            project, otherwise a new one will be created.
        """
        # Parse the filename
        if isinstance(filename, dict):
            net = filename
        else:
            filename = cls._parse_filename(filename=filename, ext='dict')
            with open(filename, mode='rb') as f:
                net = pk.load(f)

        network = GenericNetwork(project=project)
        network = cls._update_network(network=network, net=net)
        # The following list of props comes from porespy but are not good
        for item in ['pore.diameter',
                     'pore.area',
                     'throat.conduit_lengths.pore1',
                     'throat.conduit_lengths.pore2',
                     'throat.conduit_lengths.throat',
                     'throat.endpoints.head',
                     'throat.endpoints.tail',
                     'throat.diameter',
                     'throat.length',
                     'throat.volume']:
            network.pop(item, None)

        # Deal with a few other PoreSpy specific issues
        network['pore.region_volume'] = network.pop('pore.volume')

        Imported(network=network, settings=settings)

        return network.project



import pickle
import time
from openpnm.utils import Workspace, Project
from openpnm.utils import logging
from openpnm.io import GenericIO
logger = logging.getLogger(__name__)
ws = Workspace()


class OpenpnmIO(GenericIO):
    r"""
    This class contains methods used for saving and loading OpenPNM Workspaces,
    Projects, and objects.

    Notes
    -----
    The methods in this class use the ``pickle`` module from the standard
    library, which have known security issues.  Do not open '.pnm' files
    from untrusted sources.

    """

    @classmethod
    def save_object_to_file(cls, objs):
        r"""
        Saves an OpenPNM object or list of objects to a file of set of files

        Parameters
        ----------
        objs : OpenPNM Base object or list of objects
            The object(s) to be saved

        """
        if not isinstance(objs, list):
            objs = [objs]
        for item in objs:
            fname = cls._parse_filename(filename=item.name,
                                        ext=item.settings['prefix'])
            with open(fname, 'wb') as f:
                pickle.dump({item.name: item}, f)

    @classmethod
    def load_object_from_file(cls, filename, project=None):
        r"""
        Loads an OpenPNM object from a file

        Parameters
        ----------
        filename : string or path object
            The name of the file containing the object to open. Can be a
            relative or absolute path, and can be a string or path object such
            as that produced by ``pathlib``.
        project : OpenPNM Project object
            If not provided one will be created and returned by this function,
            otherwise the loaded object will be added to the given ``project``.

        Returns
        -------
        project : OpenPNM Project
            If no Project object is specified then one is created. A handle to
            the Project is returned.

        """
        if project is None:
            project = Project()
        p = cls._parse_filename(filename)
        with open(p, 'rb') as f:
            d = pickle.load(f)
        obj = project._new_object(objtype=p.suffix.strip('.'),
                                  name=p.name.split('.')[0])
        obj.update(d)
        return project

    @classmethod
    def load_object_from_dict(cls, obj, objname='', objtype='', project=None):
        r"""
        Loads a dictionary currently in memory into an OpenPNM Project

        Parameters
        ----------
        obj : dict handle
            The dictionary to load
        objname : string
            The name to give the object upon loading into the Project. If
            none is specified one will be generated.
        objtype : string
            The type of object being loaded. If not specified, than a generic
            Base class will be used.  The options are 'network', 'geometry',
            'physics', 'phase', 'algorithm'.
        project : OpenPNM Project
            A project to load the given dict into (optional)

        Returns
        -------
        project : OpenPNM Project
            If no Project object is specified then one is created. A handle to
            the Project is returned.

        """
        if project is None:
            project = Project()
        new_obj = project._new_object(objtype=objtype, name=objname)
        new_obj.update(obj)
        return new_obj

    @classmethod
    def save_project(cls, project, filename=''):
        r"""
        Save an OpenPNM Project to a file on disk

        Parameters
        ----------
        project : OpenPNM Project
            The project to save
        filename : string
            The filename to save the file
        """
        if filename == '':
            filename = project.name
        filename = cls._parse_filename(filename=filename, ext='pnm')

        # Save dictionary as pickle
        d = {project.name: project}
        with open(filename, 'wb') as f:
            pickle.dump(d, f)

    @classmethod
    def save_workspace(cls, filename=''):
        r"""
        Save the current Workspace to a file on disk

        Parameters
        ----------
        filename : string
            The filename to save the file
        """
        if filename == '':
            filename = 'workspace' + '_' + time.strftime('%Y%b%d_%H%M%p')
        filename = cls._parse_filename(filename=filename, ext='pnm')
        # Create a normal dict to store objects to prevent name errors upon
        # reopening
        d = {}
        for sim in ws.values():
            d[sim.name] = sim
        with open(filename, 'wb') as f:
            pickle.dump(d, f)

    @classmethod
    def load_workspace(cls, filename, overwrite=False):
        r"""
        Load a saved Workspace into the current one

        Parameters
        ----------
        filename : string or path object
            The name of the file to load
        overwrite : boolean
            A flag to indicate if the current Workspace should be
            overwritten when loading the new one.  The default is ``False``,
            meaning the loaded file will be added to the existing data.  Note
            that in this case Project names may clash, in which case the
            newly loaded Projects are given new names.

        Returns
        -------
        workspace : OpenPNM Workspace Object
            A handle to the Workspace, with the newly loaded Projects added
        """
        fname = cls._parse_filename(filename=filename, ext='pnm')
        temp = {}  # Read file into temporary dict
        if overwrite:
            ws.clear()
        with open(fname, 'rb') as f:
            d = pickle.load(f)
            # A normal pnm file is a dict of lists (projects)
            if isinstance(d, dict):
                for name in d.keys():
                    # If dict item is a list, assume it's a valid project
                    if isinstance(d[name], list):
                        temp[name] = d[name]
                    else:
                        raise Exception('File does not contain a valid '
                                        + 'OpenPNM Workspace')
        # Now scan through temp dict to ensure valid types and names
        conflicts = set(temp.keys()).intersection(set(ws.keys()))
        for name in list(temp.keys()):
            if name in conflicts:
                new_name = ws._gen_name()
                logger.warning('A project named ' + name + ' already exists,'
                               + ' renaming to ' + new_name)
                ws[new_name] = temp[name]
            else:
                ws[name] = temp[name]
        return ws

    @classmethod
    def load_project(cls, filename):
        r"""
        Load a saved Project file into the current Workspace

        Parameters
        ----------
        filename : string or path object
            The name of the file to load

        Returns
        -------
        project : OpenPNM Project
            A handle to the loaded Project is returned.
        """
        filename = cls._parse_filename(filename=filename, ext='pnm')
        projname = filename.name.split('.')[0]
        with open(filename, 'rb') as f:
            d = pickle.load(f)
            # A normal pnm file is a dict of lists (projects), so check first
            if type(d) is dict:
                if len(d) == 1:  # If only one project in dict
                    # Store existing keys found in Workspace
                    projects = set(ws.keys())
                    # Call load_workspace to handle file
                    temp = cls.load_workspace(filename=filename,
                                              overwrite=False)
                    # Find new addition to Workspace and return it
                    proj = list(set(temp.keys()).difference(projects))[0]
                    # Return Project handle to user and exit
                    return ws[proj]
                else:
                    raise Exception(filename.name + ' contains multiple '
                                    + 'projects, use load_workspace instead')
            # If pickle contains a single list
            elif isinstance(d, list):
                if projname not in ws.keys():
                    ws[projname] = d
                    return ws[projname]
                else:
                    newname = ws._gen_name()
                    logger.warning('Project named ' + projname
                                   + ' already present in Workspace,'
                                   + ' renaming to ' + newname)
                    ws[newname] = d
                    return ws[newname]
            else:
                raise Exception('File contents are not understood')



from xml.etree import ElementTree as ET
from flatdict import FlatDict
import numpy as np
from openpnm.io import GenericIO, Dict
from openpnm.utils import logging, Workspace

logger = logging.getLogger(__name__)
ws = Workspace()


class VTK(GenericIO):
    r"""
    The Visualization Toolkit (VTK) format defined by Kitware and used by
    Paraview
    Because OpenPNM data is unstructured, the actual output format is VTP,
    not VTK.
    """

    _TEMPLATE = """
    <?xml version="1.0" ?>
    <VTKFile byte_order="LittleEndian" type="PolyData" version="0.1">
        <PolyData>
            <Piece NumberOfLines="0" NumberOfPoints="0">
                <Points>
                </Points>
                <Lines>
                </Lines>
                <PointData>
                </PointData>
                <CellData>
                </CellData>
            </Piece>
        </PolyData>
    </VTKFile>
    """.strip()

    @classmethod
    def save(
        cls,
        network,
        phases=[],
        filename="",
        delim=" | ",
        fill_nans=None,
        fill_infs=None,
    ):
        r"""
        Save network and phase data to a single vtp file for visualizing in
        Paraview
        Parameters
        ----------
        network : OpenPNM Network Object
            The Network containing the data to be written
        phases : list, optional
            A list containing OpenPNM Phase object(s) containing data to be
            written
        filename : string, optional
            Filename to write data.  If no name is given the file is named
            after the network
        delim : string
            Specify which character is used to delimit the data names.  The
            default is ' | ' which creates a nice clean output in the Paraview
            pipeline viewer (e.g. net | property | pore | diameter)
        fill_nans : scalar
            The value to use to replace NaNs with.  The VTK file format does
            not work with NaNs, so they must be dealt with.  The default is
            `None` which means property arrays with NaNs are not written to the
            file.  Other useful options might be 0 or -1, but the user must
            be aware that these are not real values, only place holders.
        fill_infs : scalar
            The value to use to replace infs with.  The default is ``None``
            which means that property arrays containing ``None`` will *not*
            be written to the file, and a warning will be issued.  A useful
            value is
        """
        project, network, phases = cls._parse_args(network=network, phases=phases)
        # Check if any of the phases has time series
        transient = GenericIO.is_transient(phases=phases)
        if transient:
            logger.warning(
                "vtp format does not support transient data, " + "use xdmf instead"
            )
        if filename == "":
            filename = project.name
        filename = cls._parse_filename(filename=filename, ext="vtp")

        am = Dict.to_dict(
            network=network,
            phases=phases,
            interleave=True,
            categorize_by=["object", "data"],
        )
        am = FlatDict(am, delimiter=delim)
        key_list = list(sorted(am.keys()))

        network = network[0]
        points = network["pore.coords"]
        pairs = network["throat.conns"]
        num_points = np.shape(points)[0]
        num_throats = np.shape(pairs)[0]

        root = ET.fromstring(VTK._TEMPLATE)
        piece_node = root.find("PolyData").find("Piece")
        piece_node.set("NumberOfPoints", str(num_points))
        piece_node.set("NumberOfLines", str(num_throats))
        points_node = piece_node.find("Points")
        coords = VTK._array_to_element("coords", points.T.ravel("F"), n=3)
        points_node.append(coords)
        lines_node = piece_node.find("Lines")
        connectivity = VTK._array_to_element("connectivity", pairs)
        lines_node.append(connectivity)
        offsets = VTK._array_to_element("offsets", 2 * np.arange(len(pairs)) + 2)
        lines_node.append(offsets)

        point_data_node = piece_node.find("PointData")
        cell_data_node = piece_node.find("CellData")
        for key in key_list:
            array = am[key]
            if array.dtype == "O":
                logger.warning(key + " has dtype object," + " will not write to file")
            else:
                if array.dtype == np.bool:
                    array = array.astype(int)
                if np.any(np.isnan(array)):
                    if fill_nans is None:
                        logger.warning(key + " has nans," + " will not write to file")
                        continue
                    else:
                        array[np.isnan(array)] = fill_nans
                if np.any(np.isinf(array)):
                    if fill_infs is None:
                        logger.warning(key + " has infs," + " will not write to file")
                        continue
                    else:
                        array[np.isinf(array)] = fill_infs
                element = VTK._array_to_element(key, array)
                if array.size == num_points:
                    point_data_node.append(element)
                elif array.size == num_throats:
                    cell_data_node.append(element)

        tree = ET.ElementTree(root)
        tree.write(filename)

        with open(filename, "r+") as f:
            string = f.read()
            string = string.replace("</DataArray>", "</DataArray>\n\t\t\t")
            f.seek(0)
            # consider adding header: '<?xml version="1.0"?>\n'+
            f.write(string)

    @classmethod
    def load(cls, filename, project=None, delim=" | "):
        r"""
        Read in pore and throat data from a saved VTK file.
        Parameters
        ----------
        filename : string (optional)
            The name of the file containing the data to import.  The formatting
            of this file is outlined below.
        project : OpenPNM Project object
            A GenericNetwork is created and added to the specified Project.
            If no Project is supplied then one will be created and returned.
        """
        net = {}

        filename = cls._parse_filename(filename, ext="vtp")
        tree = ET.parse(filename)
        piece_node = tree.find("PolyData").find("Piece")

        # Extract connectivity
        conn_element = piece_node.find("Lines").find("DataArray")
        conns = VTK._element_to_array(conn_element, 2)
        # Extract coordinates
        coord_element = piece_node.find("Points").find("DataArray")
        coords = VTK._element_to_array(coord_element, 3)

        # Extract pore data
        for item in piece_node.find("PointData").iter("DataArray"):
            key = item.get("Name")
            array = VTK._element_to_array(item)
            net[key] = array
        # Extract throat data
        for item in piece_node.find("CellData").iter("DataArray"):
            key = item.get("Name")
            array = VTK._element_to_array(item)
            net[key] = array

        if project is None:
            project = ws.new_project()
        project = Dict.from_dict(dct=net, project=project, delim=delim)

        # Clean up data values, if necessary, like convert array's of
        # 1's and 0's into boolean.
        project = cls._convert_data(project)

        # Add coords and conns to network
        network = project.network
        network.update({"throat.conns": conns})
        network.update({"pore.coords": coords})

        return project

    @classmethod
    def _array_to_element(cls, name, array, n=1):
        dtype_map = {
            "int8": "Int8",
            "int16": "Int16",
            "int32": "Int32",
            "int64": "Int64",
            "uint8": "UInt8",
            "uint16": "UInt16",
            "uint32": "UInt32",
            "uint64": "UInt64",
            "float32": "Float32",
            "float64": "Float64",
            "str": "String",
        }
        element = None
        if str(array.dtype) in dtype_map.keys():
            element = ET.Element("DataArray")
            element.set("Name", name)
            element.set("NumberOfComponents", str(n))
            element.set("type", dtype_map[str(array.dtype)])
            element.text = "\t".join(map(str, array.ravel()))
        return element

    @classmethod
    def _element_to_array(cls, element, n=1):
        string = element.text
        dtype = element.get("type")
        array = np.fromstring(string, sep="\t")
        array = array.astype(dtype)
        if n != 1:
            array = array.reshape(array.size // n, n)
        return array



# -*- coding: utf-8 -*-
"""
===============================================================================
MixedInvasionPercolationCoop: IP allowing pores and throats to invade separately
With added cooperative filling algorithms
===============================================================================

"""
import time
import logging
import heapq as hq
import scipy as sp
import numpy as np
from scipy.sparse import coo_matrix, dok_matrix
from openpnm.algorithms import MixedInvasionPercolation
from transforms3d._gohlketransforms import angle_between_vectors

logger = logging.getLogger(__name__)


class MixedInvasionPercolationCoop(MixedInvasionPercolation):
    r"""
    An implemetation of invasion percolation which can invade bonds, sites or a
    mixture of both. Inlets can be treated as individual injection points that
    share a common pressure or have their own and progess independently.
    Inlets can also be single pores or clusters.

    Parameters
    ----------
    network : OpenPNM Network object
        The Network upon which the invasion should occur.

    Notes
    ----
    n/a

    """

    def __init__(self, settings={}, **kwargs):
        def_set = {
            "pore_entry_pressure": "pore.entry_pressure",
            "throat_entry_pressure": "throat.entry_pressure",
            "snap_off": "",
            "invade_isolated_Ts": False,
            "late_pore_filling": "",
            "late_throat_filling": "",
            "gui": {
                "setup": {
                    "pore_entry_pressure": "",
                    "throat_entry_pressure": "",
                    "snap_off": "",
                    "invade_isolated_Ts": "",
                },
                "set_inlets": {"pores": None, "clusters": None},
                "set_outlets": {"pores": None, "overwrite": False},
                "apply_flow": {"flowrate": None},
                "apply_trapping": {"partial": False},
                "set_residual": {"pores": None, "overwrite": False},
            },
        }
        super().__init__(**kwargs)
        self.settings.update(def_set)
        self.settings.update(settings)

    def setup(
        self,
        phase=None,
        pore_entry_pressure="pore.entry_pressure",
        throat_entry_pressure="throat.entry_pressure",
        snap_off="",
        invade_isolated_Ts=False,
        late_pore_filling="",
        late_throat_filling="",
        cooperative_pore_filling="",
    ):
        r"""
        Used to specify necessary arguments to the simulation.  This method is
        useful for resetting the algorithm or applying more explicit control.

        Parameters
        ----------
        phase : OpenPNM Phase object
            The Phase object containing the physical properties of the invading
            fluid.

        pore_entry_pressure : string
            The dictionary key on the Phase object where the pore entry
            pressure values are stored.  The default is
            'pore.entry_pressure'.

        throat_entry_pressure : string
            The dictionary key on the Phase object where the throat entry
            pressure values are stored.  The default is
            'throat.entry_pressure'.

        snap_off : string
            The dictionary key on the Phase object where the throat snap-off
            pressure values are stored.

        invade_isolated_Ts : boolean
            If True, isolated throats are invaded at the higher invasion
            pressure of their connected pores.

        late_pore_filling : string
            The name of the model used to determine late pore filling as
            a function of applied pressure.

        late_throat_filling : string
            The name of the model used to determine late throat filling as
            a function of applied pressure.

        cooperative_pore_filling : string
            The name of the model used to determine the meniscus properties
            required for assessing cooperative pore filling.
        """
        if phase:
            self.settings["phase"] = phase.name
        if throat_entry_pressure:
            self.settings["throat_entry_pressure"] = throat_entry_pressure
            phase = self.project.find_phase(self)
        self["throat.entry_pressure"] = phase[self.settings["throat_entry_pressure"]]
        if len(np.shape(self["throat.entry_pressure"])) > 1:
            self._bidirectional = True
        else:
            self._bidirectional = False
        if pore_entry_pressure:
            self.settings["pore_entry_pressure"] = pore_entry_pressure
            phase = self.project.find_phase(self)
        self["pore.entry_pressure"] = phase[self.settings["pore_entry_pressure"]]
        if snap_off:
            self.settings["snap_off"] = snap_off
        if invade_isolated_Ts:
            self.settings["invade_isolated_Ts"] = invade_isolated_Ts
        if late_pore_filling:
            self.settings["late_pore_filling"] = late_pore_filling
        if late_throat_filling:
            self.settings["late_throat_filling"] = late_throat_filling
        if cooperative_pore_filling:
            self.settings["cooperative_pore_filling"] = cooperative_pore_filling
        self.reset()

    def _max_pressure(self):
        phase = self.project.find_phase(self)
        if self.settings["throat_entry_pressure"]:
            max_tPc = np.max(phase[self.settings["throat_entry_pressure"]])
        else:
            max_tPc = 0.0
        if self.settings["pore_entry_pressure"]:
            max_pPc = np.max(phase[self.settings["pore_entry_pressure"]])
        else:
            max_pPc = 0.0
        return np.max([max_tPc, max_pPc])

    def _my_dot(self, a, b):
        return a[:, 0] * b[:, 0] + a[:, 1] * b[:, 1] + a[:, 2] * b[:, 2]

    def trilaterate_v(self, P1, P2, P3, r1, r2, r3):
        r"""
        Find whether 3 spheres intersect
        """
        temp1 = P2 - P1
        e_x = temp1 / np.linalg.norm(temp1, axis=1)[:, np.newaxis]
        temp2 = P3 - P1
        i = self._my_dot(e_x, temp2)[:, np.newaxis]
        temp3 = temp2 - i * e_x
        e_y = temp3 / np.linalg.norm(temp3, axis=1)[:, np.newaxis]
        d = np.linalg.norm(P2 - P1, axis=1)[:, np.newaxis]
        j = self._my_dot(e_y, temp2)[:, np.newaxis]
        x = (r1 * r1 - r2 * r2 + d * d) / (2 * d)
        y = (r1 * r1 - r3 * r3 - 2 * i * x + (i * i) + (j * j)) / (2 * j)
        temp4 = r1 * r1 - x * x - y * y
        return temp4 >= 0

    def _get_throat_pairs(self):
        r"""
        Generate an array of pores with all connected throats and pairs of
        throats that connect to the same pore
        """
        network = self.project.network
        # Collect all throat pairs sharing a pore as list of lists
        neighbor_Ts = network.find_neighbor_throats(
            pores=network.pores(), flatten=False
        )
        # Pores associated to throat
        Ps = []
        # Throats connected to each pore
        Ts = []
        # Pairs of throats sharing a pore
        T1 = []
        T2 = []
        start = 0
        # Build lookup pair index arrays for each coordination number up to the
        # Maximum coordination max_c
        max_c = np.amax(network.num_neighbors(pores=network.Ps, flatten=False))
        pair_T1 = []
        pair_T2 = []
        logger.info("Building throat pair matrices")
        for num_t in range(max_c + 1):
            temp1 = []
            temp2 = []
            for t1 in range(num_t)[:-1]:
                for t2 in range(num_t)[t1 + 1 :]:
                    temp1.append(t1)
                    temp2.append(t2)
            pair_T1.append(np.asarray(temp1))
            pair_T2.append(np.asarray(temp2))
        for p, nTs in enumerate(neighbor_Ts):
            num_t = len(nTs)
            for i in range(num_t):
                Ps.append(p)
                Ts.append(nTs[i])
            # Pair indices into Ts
            tempt1 = pair_T1[num_t] + start
            tempt2 = pair_T2[num_t] + start
            for i in range(len(tempt1)):
                T1.append(tempt1[i])
                T2.append(tempt2[i])
            start += num_t
        # Nt * 2 long
        Ps = np.asarray(Ps)
        Ts = np.asarray(Ts)
        # indices into the above arrays based on throat pairs
        T1 = np.asarray(T1)
        T2 = np.asarray(T2)

        return Ps, Ts, T1, T2

    def _apply_cen_to_throats(self, p_cen, t_cen, t_norm, men_cen):
        r"""
        Take the pore center and throat center and work out which way
        the throat normal is pointing relative to the vector between centers.
        Offset the meniscus center along the throat vector in the correct
        direction
        """
        v = p_cen - t_cen
        sign = np.sign(np.sum(v * t_norm, axis=1))
        c3 = np.vstack((men_cen * sign, men_cen * sign, men_cen * sign)).T
        coords = t_cen + c3 * t_norm
        return coords

    def _transform_point_normal(self, point, normal):
        r"""
        Transforms point normal plane definition to parametric form
        Ax + By +Cz + D = 0
        """
        return [
            a
            for a in zip(
                normal[:, 0], normal[:, 1], normal[:, 2], -self._my_dot(point, normal)
            )
        ]

    def _plane_intersect(self, a, b):
        """
        a, b   4-tuples/lists
               Ax + By +Cz + D = 0
               A, B, C, D in order
        output: 2 points on line of intersection, np.arrays, shape (3,)
        https://bit.ly/2LkBEyc
        """
        a_vec, b_vec = np.array(a[:3]), np.array(b[:3])
        aXb_vec = np.cross(a_vec, b_vec)
        A = np.array([a_vec, b_vec, aXb_vec])
        d = np.array([-a[3], -b[3], 0.0]).reshape(3, 1)
        if np.linalg.det(A) == 0:
            return None, None
        else:
            p_inter = np.linalg.solve(A, d).T
            return p_inter[0], (p_inter + aXb_vec)[0]

    def _t(self, p, q, r):
        r"""
        Equation of line passing through points p and q parameterized by t
        https://bit.ly/2EpQ6DD
        """
        x = p - q
        return np.dot(r - q, x) / np.dot(x, x)

    def _distance(self, p, q, r):
        r"""
        Shortest distance between line passing through p and q and point r
        https://bit.ly/2EpQ6DD
        """
        return np.linalg.norm(self._t(p, q, r) * (p - q) + q - r)

    def _pair_permutations(self, n):
        perms = []
        for i in range(n):
            for j in range(n - (i + 1)):
                perms.append([i, j + i + 1])
        return perms

    def _perpendicular_vector(self, v, v_ref=None):
        if v_ref is None:
            np.array([1.0, 0.0, 0.0])
        return np.cross(v, v_ref)

    def _throat_pair_angle(self, t1, t2, pore, network):
        conn1 = network["throat.conns"][t1]
        conn2 = network["throat.conns"][t2]
        lookup = np.vstack((pore, pore)).T
        p1 = conn1[conn1 != lookup]
        p2 = conn2[conn2 != lookup]
        v1 = network["pore.coords"][pore] - network["pore.coords"][p1]
        v2 = network["pore.coords"][pore] - network["pore.coords"][p2]
        v_ref = self._perpendicular_vector(v1, v2)
        v1_perp = self._perpendicular_vector(v1, v_ref)
        v2_perp = self._perpendicular_vector(v2, v_ref)
        return angle_between_vectors(v1_perp, v2_perp, axis=1)

    def setup_coop_filling(self, inv_points=None):
        r"""
        Populate the coop filling throat-throat pair matrix
        """
        self._setup_coop_filling_creep(inv_points)
        self._setup_coop_filling_bulge(inv_points)

    def _setup_coop_filling_creep(self, inv_points=None):
        r"""
        This coop filling model iterates through the invasion pressures set by
        the inv_points variable and calls the meniscus model whose dictionary
        key must be given in the algorithm's setup.
        The meniscus model supplies the position of the meniscus inside each
        throat as if there were a meniscus in every throat for a given pressure
        The contact line of the meniscus traces a circle around the inner
        surface of the throat which is assumed to be toroidal.
        The contact circle lies on a plane that is defined by the throat's
        normal vector which runs along the axis of symmetry of the torus.
        For every pore, every connecting throat is compared with each of it's
        neighboring throats connected to the same pore. If the planes intersect
        then the meniscus contact circles may eventually touch if they can
        advance enough. For highly wetting fluid the contact point may be
        advanced well into the throat whilst still being at negative capillary
        pressure.
        """
        start = time.time()
        net = self.project.network
        phase = self.project.find_phase(self)
        all_phys = self.project.find_physics(phase=phase)
        if inv_points is None:
            inv_points = np.arange(0, 1.01, 0.01) * self._max_pressure()
        # Throat centroids
        try:
            t_centroids = net["throat.centroid"]
        except KeyError:
            t_centroids = np.mean(net["pore.coords"][net["throat.conns"]], axis=1)
        # Throat normal vector
        try:
            t_norms = net["throat.normal"]
        except KeyError:
            t_norms = (
                net["pore.coords"][net["throat.conns"][:, 1]]
                - net["pore.coords"][net["throat.conns"][:, 0]]
            )
        cpf = self.settings["cooperative_pore_filling"]
        model = all_phys[0].models[cpf]
        # Throat Diameter and fiber radius
        try:
            t_rad = net[model["throat_diameter"]] / 2 + model["r_toroid"]
        except KeyError:
            t_rad = net["throat.diameter"] / 2
        # Equations of throat planes at the center of each throat
        planes = self._transform_point_normal(t_centroids, t_norms)
        Nt = net.Nt
        adj_mat = dok_matrix((Nt, Nt), dtype=int)
        # Run through and build the throat-throat pair adjacency matrix
        # If planes of throats intersect then meniscii in throats may also
        # Intersect at a given pressure.
        for pore in net.Ps:
            ts = net.find_neighbor_throats(pores=pore)
            perms = self._pair_permutations(len(ts))
            for perm in perms:
                ta, tb = ts[perm]
                p, q = self._plane_intersect(planes[ta], planes[tb])
                if p is not None:
                    d1 = self._distance(p, q, t_centroids[ta])
                    d2 = self._distance(p, q, t_centroids[tb])
                    if (t_rad[ta] >= d1) and (t_rad[tb] >= d2):
                        adj_mat[ta, tb] = pore + 1
        # Meniscus Filling Angle
        tfill_angle = cpf + ".alpha"
        # Capillary pressure adjacency maxtrix
        pairs = np.asarray([list(key) for key in adj_mat.keys()])
        pores = np.asarray([adj_mat[key] for key in adj_mat.keys()]) - 1
        self.tt_Pc = adj_mat.copy().astype(float).tocsr()
        # Initialize the pressure matrix with nans
        # This is used to check for the first intersection pressure and
        # Prevent overwriting
        self.tt_Pc[pairs[:, 0], pairs[:, 1]] = np.nan
        angles = self._throat_pair_angle(pairs[:, 0], pairs[:, 1], pores, net)
        T1 = pairs[:, 0]
        T2 = pairs[:, 1]
        hits = []
        for Pc in inv_points:
            # Don't use zero as can get strange numbers in menisci data
            if Pc == 0.0:
                Pc = 1e-6
            # regenerate model with new target Pc
            for phys in all_phys:
                phys.models[cpf]["target_Pc"] = Pc
                phys.regenerate_models(propnames=cpf)
            # check whether this throat pair already has a coop value
            check_nans = np.asarray(np.isnan(self.tt_Pc[T1, T2]).tolist()[0])
            fill_angle_sum = np.sum(phase[tfill_angle][pairs], axis=1)
            coalescence = fill_angle_sum >= angles
            mask = check_nans * coalescence
            if np.any(mask):
                self.tt_Pc[T1[mask], T2[mask]] = Pc
                hits.append(Pc)
        # Change to lil for single throat lookups
        #        self.tt_Pc = self.tt_Pc.tolil()
        logger.info(
            "Coop filling finished in " + str(np.around(time.time() - start, 2)) + " s"
        )
        logger.info("Coop Hits", np.unique(np.asarray(hits)))

    def _setup_coop_filling_bulge(self, inv_points=None):
        r"""
        Evaluate the cooperative pore filling condition that the combined
        filling angle in next neighbor throats cannot exceed the geometric
        angle between their throat planes.
        This is used when the invading fluid has access to multiple throats
        connected to a pore

        Parameters
        ----------
        inv_points : array_like
            The invasion pressures at which to assess coopertive pore filling.
        """
        net = self.project.network
        phase = self.project.find_phase(self)
        all_phys = self.project.find_physics(phase=phase)
        if inv_points is None:
            inv_points = np.arange(0, 1.01, 0.01) * self._max_pressure()

        start = time.time()
        cpf = self.settings["cooperative_pore_filling"]
        tfill_angle = cpf + ".alpha"
        tmen_rad = cpf + ".radius"
        tmen_cen = cpf + ".center"
        try:
            # The following properties will all be there for Voronoi
            p_centroids = net["pore.centroid"]
            t_centroids = net["throat.centroid"]
            p_rad = net["pore.indiameter"] / 2
            t_norms = net["throat.normal"]
        except KeyError:
            # Chances are this isn't Voronoi so calculate or replace all
            p_centroids = net["pore.coords"]
            temp = net["pore.coords"][net["throat.conns"]]
            t_centroids = np.mean(temp, axis=1)
            p_rad = net["pore.diameter"] / 2
            t_norms = net["throat.normal"]

        Ps, Ts, T1, T2 = self._get_throat_pairs()
        # Network indices for the pairs
        pps = Ps[T1]
        pt1 = Ts[T1]
        pt2 = Ts[T2]
        # Arrays to get upper and lower triangular for reference to later
        # When checking for phase presence one throat at a time - need to
        # Check all throat pairs
        T_all_t1 = Ts[np.concatenate((T1, T2))]
        # T_all_t2 = Ts[np.concatenate((T2, T1))]
        # Throat-Throat cooperative filling pressure
        data = np.ones(len(T_all_t1), dtype=float)
        data.fill(np.nan)
        # coo --> useful for building matrix
        # coo = coo_matrix(
        #     (data, (T_all_t1, T_all_t2)), shape=(len(T_all_t1), len(T_all_t2))
        # )
        # csr --> for indexing into entire matrix when populating with
        # actual coop filling values
        #        self.tt_Pc = coo.tocsr()
        # Pair pore center and radius
        pp_cen = p_centroids[pps]
        pp_rad = p_rad[pps]
        # Make sure throat normals are unit vector
        unit = np.linalg.norm(t_norms, axis=1)
        t_norms /= np.vstack((unit, unit, unit)).T

        for Pc in inv_points:
            # regenerate model with new target Pc
            for phys in all_phys:
                phys.models[cpf]["target_Pc"] = Pc
                phys.regenerate_models(propnames=cpf)
            men_cen_dist = phase[tmen_cen]
            # Work out meniscii coord for each direction along the throat
            men_cen_coord = self._apply_cen_to_throats(
                p_centroids[Ps], t_centroids[Ts], t_norms[Ts], men_cen_dist[Ts]
            )
            # Pair centers
            pc1 = men_cen_coord[T1]
            pc2 = men_cen_coord[T2]
            # Center to center vector between neighboring meniscii
            c2c = pc1 - pc2
            dist = np.linalg.norm(c2c, axis=1)
            # Pair meniscii radii
            pr1 = phase[tmen_rad][pt1]
            pr2 = phase[tmen_rad][pt2]
            # nans may exist if pressure is outside the range
            # set these to zero to be ignored by next step without
            # causing RuntimeWarning
            pr1[np.isnan(pr1)] = 0
            pr2[np.isnan(pr2)] = 0
            # Negative mensicii radii means positive pressure
            # Assume meniscii only interact when bulging into pore
            check_neg = np.logical_and(pr1 < 0, pr2 < 0)
            # simple initial distance check on sphere rads
            check_rads = (np.abs(pr1 + pr2)) >= dist
            # check whether the filling angle is ok at this Pc
            check_alpha_T1 = ~np.isnan(phase[tfill_angle][pt1])
            check_alpha_T2 = ~np.isnan(phase[tfill_angle][pt2])
            check_alpha = check_alpha_T1 * check_alpha_T2
            # check whether this throat pair already has a coop value
            check_nans = np.asarray(np.isnan(self.tt_Pc[pt1, pt2]).tolist()[0])
            mask = check_neg * check_alpha * check_nans * check_rads
            # if all checks pass
            if np.any(mask):
                # Check if intersecting circle lies within pore
                inter = self.trilaterate_v(
                    P1=pc1[mask],
                    P2=pc2[mask],
                    P3=pp_cen[mask],
                    r1=pr1[mask][:, np.newaxis],
                    r2=pr2[mask][:, np.newaxis],
                    r3=pp_rad[mask][:, np.newaxis],
                )
                inter = inter.flatten()
                if np.any(inter):
                    self.tt_Pc[pt1[mask][inter], pt2[mask][inter]] = Pc
                    self.tt_Pc[pt2[mask][inter], pt1[mask][inter]] = Pc
            # Save meniscii data for each throat into each connecting pore
            men_data = {}
            men_data["Ps"] = Ps
            men_data["Ts"] = Ts
            men_data["cen"] = men_cen_coord
            men_data["rad"] = phase[tmen_rad][Ts]

        # Change to lil for single throat lookups
        self.tt_Pc = self.tt_Pc.tolil()
        logger.info(
            "Coop filling finished in " + str(np.around(time.time() - start, 2)) + " s"
        )

    def _check_coop(self, pore, queue):
        r"""
        Method run in loop after every pore invasion. All connecting throats
        are now given access to the invading phase. Two throats with access to
        the invading phase can cooperatively fill any pores that they are both
        connected to, common pores.
        The invasion of theses throats connected to the common pore is handled
        elsewhere.
        """
        net = self.project.network
        t_inv = "throat.invasion_sequence"
        p_inv = "pore.invasion_sequence"
        for throat in net.find_neighbor_throats(pores=pore):
            # A pore has just been invaded, all it's throats now have
            # An interface residing inside them
            if self[t_inv][throat] == -1:
                # If the throat is not the invading throat that gave access
                # to this pore, get the pores that this throat connects with
                a = set(net["throat.conns"][throat])
                # Get a list of pre-calculated coop filling pressures for all
                # Throats this throat can coop fill with
                ts_Pc = self.tt_Pc.data[throat]
                # Network indices of throats that can act as filling pairs
                ts = self.tt_Pc.rows[throat]
                # If there are any potential coop filling throats
                if np.any(~np.isnan(ts_Pc)):
                    ts_Pc = np.asarray(ts_Pc)
                    ts = np.asarray(ts)
                    ts = ts[~np.isnan(ts_Pc)]
                    ts_Pc = ts_Pc[~np.isnan(ts_Pc)]
                    # For each throat find the common pore and the uncommon
                    # pores
                    for i, t in enumerate(ts):
                        # Find common pore (cP) and uncommon pores (uPs)
                        b = set(net["throat.conns"][t])
                        cP = list(a.intersection(b))
                        uPs = list(a.symmetric_difference(b))
                        # If the common pore is not invaded but the others are
                        # The potential coop filling event can now happen
                        # Add the coop pressure to the queue
                        if (np.all(self[p_inv][uPs] > -1)) and (self[p_inv][cP] == -1):
                            # Coop pore filling fills the common pore
                            # The throats that gave access are not invaded now
                            # However, isolated throats between invaded pores
                            # Are taken care of elsewhere...
                            hq.heappush(queue, [ts_Pc[i], list(cP), "pore"])



r"""

.. autofunction:: openpnm.models.phases.density.water
.. autofunction:: openpnm.models.phases.density.standard
.. autofunction:: openpnm.models.phases.density.ideal_gas

"""


def standard(target, mol_weight='pore.molecular_weight',
             molar_density='pore.molar_density'):
    r"""
    Calculates the mass density from the molecular weight and molar density

    Parameters
    ----------
    target : OpenPNM Object
        The object for which these values are being calculated.  This
        controls the length of the calculated array, and also provides
        access to other necessary thermofluid properties.

    mol_weight : string
        The dictionary key containing the molecular weight values (kg/mol)

    molar_density : string
        The dictionary key containing the molar density values (mol/m3)

    Returns
    -------
    value : NumPy ndarray
        Array containing density values (kg/m3)

    """
    MW = target[mol_weight]
    rho = target[molar_density]
    value = rho*MW
    return value


def ideal_gas(target, pressure='pore.pressure', temperature='pore.temperature',
              mol_weight='pore.molecular_weight'):
    r"""
    Uses ideal gas law to calculate the mass density of an ideal gas

    Parameters
    ----------
    target : OpenPNM Object
        The object for which these values are being calculated.  This
        controls the length of the calculated array, and also provides
        access to other necessary thermofluid properties.

    pressure : string
        The dictionary key containing the pressure values (Pa)

    temperature : string
        The dictionary key containing the temperature values (K)

    mol_weight : string
        The dictionary key containing the molecular weight values (kg/mol)

    Returns
    -------
    value : NumPy ndarray
        Array containing density values in [kg/m3]

    """

    P = target[pressure]
    T = target[temperature]
    MW = target[mol_weight]
    R = 8.31447
    value = P/(R*T)*MW
    return value


def water(target, temperature='pore.temperature', salinity='pore.salinity'):
    r"""
    Calculates density of pure water or seawater at atmospheric pressure
    using Eq. (8) given by Sharqawy et. al [1]. Values at temperature higher
    than the normal boiling temperature are calculated at the saturation
    pressure.

    Parameters
    ----------
    target : OpenPNM Object
        The object for which these values are being calculated.  This
        controls the length of the calculated array, and also provides
        access to other necessary thermofluid properties.

    temperature : string
        The dictionary key containing the temperature values.  Temperature must
        be in Kelvin for this emperical equation to work

    salinity : string
        The dictionary key containing the salinity values.  Salinity must be
        expressed in g of salt per kg of solution (ppt).


    Returns
    -------
    value : NumPy ndarray
        The density of water/seawater in [kg/m3]

    Notes
    -----
     T must be in K, and S in g of salt per kg of phase, or ppt (parts per
        thousand)
    VALIDITY: 273 < T < 453 K; 0 < S < 160 g/kg;
    ACCURACY: 0.1 %

    References
    ----------
    [1] Sharqawy M. H., Lienhard J. H., and Zubair, S. M., Desalination and
    Water Treatment, 2010.

    """
    T = target[temperature]
    if salinity in target.keys():
        S = target[salinity]
    else:
        S = 0
    a1 = 9.9992293295E+02
    a2 = 2.0341179217E-02
    a3 = -6.1624591598E-03
    a4 = 2.2614664708E-05
    a5 = -4.6570659168E-08
    b1 = 8.0200240891E-01
    b2 = -2.0005183488E-03
    b3 = 1.6771024982E-05
    b4 = -3.0600536746E-08
    b5 = -1.6132224742E-11
    TC = T-273.15
    rho_w = a1 + a2*TC + a3*TC**2 + a4*TC**3 + a5*TC**4
    d_rho = b1*S + b2*S*TC + b3*S*(TC**2) + b4*S*(TC**3) + b5*(S**2)*(TC**2)
    rho_sw = rho_w + d_rho
    value = rho_sw
    return value



from openpnm.core import Base
import numpy as np


class Subdomain(Base):
    r"""
    This subclass of the Base class provides the ability assign the object
    to specific locations (pores and throats) in the domain.  This class
    is subclassed by GenericGeometry and GenericPhysics.

    Notes
    -----
    The following table list the two methods added to Base by this subclass.

    +---------------------+---------------------------------------------------+
    | Methods             | Description                                       |
    +=====================+===================================================+
    | ``_add_locations``  | Specified which pores and throats the object      |
    |                     | should be assigned to                             |
    +---------------------+---------------------------------------------------+
    | ``_drop_locations`` | Removes the object from the specified pores and   |
    |                     | throats                                           |
    +---------------------+---------------------------------------------------+
    | ``_set_locations``  | The actual general method called by the above    |
    +---------------------+---------------------------------------------------+

    The Project object has two methods, ``check_geometry_health`` and
    ``check_physics_health`` that look to make sure all locations are
    assigned to one and only one Geometry and/or Physics.

    """

    def __getitem__(self, key):
        element = key.split('.')[0]
        # Find boss object (either phase or network)
        boss = self.project.find_full_domain(self)
        # Try to get vals directly first
        vals = self.get(key)
        if vals is None:  # Otherwise invoke search
            inds = boss._get_indices(element=element, labels=self.name)
            try:  # Will invoke interleave data if necessary
                vals = boss[key]  # Will return nested dict if present
                if type(vals) is dict:  # Index into each array in nested dict
                    for item in vals:
                        vals[item] = vals[item][inds]
                else:  # Otherwise index into single array
                    vals = vals[inds]
            except KeyError:
                vals = super().__getitem__(key)
        return vals

    def __setitem__(self, key, value):
        # If value is a dict, skip all this.  The super class will parse
        # the dict individually, at which point the below is called.
        if self.project and not hasattr(value, 'keys'):
            proj = self.project
            boss = proj.find_full_domain(self)
            keys = boss.keys(mode='all', deep=True)
            # Prevent 'pore.foo' on subdomain when already present on boss
            if key in set(boss.keys()).difference(set(self.keys())):
                hit = [i for i in keys if i.startswith(key)][0]
                raise Exception('Cannot create ' + key + ' when '
                                + hit + ' is already defined')
        super().__setitem__(key, value)

    def _add_locations(self, pores=[], throats=[]):
        r"""
        Adds associations between an object and its boss object at the
        given pore and/or throat locations.

        Parameters
        ----------
        pores and throats : array_like
            The pore and/or throat locations for which the association should
            be added.  These indices are for the full domain.

        Notes
        -----
        For *Physics* objects, the boss is the *Phase* with which it was
        assigned, while for *Geometry* objects the boss is the *Network*.

        """
        boss = self.project.find_full_domain(self)
        pores = boss._parse_indices(pores)
        throats = boss._parse_indices(throats)
        if len(pores) > 0:
            self._set_locations(element='pore', indices=pores, mode='add')
        if len(throats) > 0:
            self._set_locations(element='throat', indices=throats, mode='add')

    def _drop_locations(self, pores=[], throats=[]):
        r"""
        Removes association between an objectx and its boss object at the
        given pore and/or throat locations.

        Parameters
        ----------
        pores and throats : array_like
            The pore and/or throat locations from which the association should
            be removed.  These indices refer to the full domain.

        Notes
        -----
        For *Physics* objects, the boss is the *Phase* with which it was
        assigned, while for *Geometry* objects the boss is the *Network*.

        """
        boss = self.project.find_full_domain(self)
        pores = boss._parse_indices(pores)
        throats = boss._parse_indices(throats)
        if len(pores) > 0:
            self._set_locations(element='pore', indices=pores, mode='drop')
        if len(throats) > 0:
            self._set_locations(element='throat', indices=throats, mode='drop')

    def _set_locations(self, element, indices, mode):
        r"""
        This private method is called by ``set_locations`` and
        ``remove_locations`` as needed.

        """
        boss = self.project.find_full_domain(self)
        element = self._parse_element(element=element, single=True)

        # Make sure label array exists in boss
        if (element+'.'+self.name) not in boss.keys():
            boss[element+'.'+self.name] = False

        # Check to ensure indices aren't already assigned
        if mode == 'add':
            if self._isa('geometry'):
                objs = self.project.geometries().keys()
            else:
                objs = self.project.physics().keys()
            for name in objs:
                if element+'.'+name in boss.keys():
                    if np.any(boss[element+'.'+name][indices]):
                        raise Exception('Given indices are already assigned to' + name)

        # Find mask of existing locations (network indexing)
        mask = boss[element+'.'+self.name]
        # Update mask with new locations (either add or remove)
        if mode == 'add':
            mask = mask + boss._tomask(indices=indices, element=element)
        elif mode == 'drop':
            mask = mask * (~boss._tomask(indices=indices, element=element))
        # Change size of all arrays on self
        for item in self.keys(element=element, mode='all'):
            self.update({item: boss[item][mask]})
        # Update label array in network
        boss[element+'.'+self.name] = mask
        # Remove label from boss if ALL locations are removed
        if mode == 'drop':
            if ~np.any(boss[element+'.'+self.name]):
                boss[element+'.'+self.name] = False



import numpy as np
from openpnm.utils import logging, Project, Workspace, PrintableDict
from openpnm.phases import GenericPhase
from openpnm.physics import GenericPhysics
from openpnm.algorithms import FickianDiffusion
from openpnm.algorithms.metrics import GenericMetric
from openpnm import models
from openpnm import topotools
logger = logging.getLogger(__name__)
ws = Workspace()


class FormationFactor(GenericMetric):
    r"""
    This class works by applying 'value' boundary conditions across the
    domain to find molar flow, then using Fick's law to back-calculate
    the effective diffusivity of the domain.  The formation factor is
    defined as:

    .. math::

        F = \frac{D_{AB}}{D_{eff}} > 1

    where

    .. math::

        D_{eff} = \frac{n_{A} L }{\Delta C_{A} A }

    and

    .. math::

        D_{eff} = D_{AB} \frac{\varepsilon}{\tau}

    The formation factor is a convenient metric to compare diffusion in
    different pore networks since it does not require knowledge of the network
    porosity, unlike tortuosity. The porosity of a pore network is difficult
    to determine, mainly because the bulk volume of a network is not
    well known.

    Examples
    --------
    >>> import openpnm as op
    >>> pn = op.network.Cubic(shape=[10, 10, 10], spacing=1e-5)
    >>> geo = op.geometry.StickAndBall(network=pn)

    Now find the formation factor of the network:

    >>> F = op.algorithms.metrics.FormationFactor(network=pn)
    >>> F.run()
    >>> print(F.results)
    ――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――
    Direction                           Formation Factor
    ――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――
    x                                   ...
    y                                   ...
    z                                   ...
    ――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――――

    """

    def __init__(self, network=None, project=None, settings={}, **kwargs):
        self.settings.update({'inlets': {'x': 'left',
                                         'y': 'front',
                                         'z': 'top'},
                              'outlets': {'x': 'right',
                                          'y': 'back',
                                          'z': 'bottom'},
                              'areas': {'x': None,
                                        'y': None,
                                        'z': None},
                              'lengths': {'x': None,
                                          'y': None,
                                          'z': None}})

        self.results = PrintableDict()
        self.results._value = "Formation Factor"
        self.results._key = "Direction"
        if network is None:
            network = project.network
        if project is None:
            project = network.project
#        project = ws.copy_project(network.project)
#        keep = [network] + list(project.geometries().values())
#        for i in project:
#            if i not in keep:
#                project.purge_object(i)
        super().__init__(network=network, project=project, **kwargs)

    def run(self):
        r"""
        Execute the diffusion simulations in the principle directions.

        """
        phase = GenericPhase(network=self.network)
        phase['pore.diffusivity'] = 1.0
        phase['throat.diffusivity'] = 1.0
        mod = models.physics.diffusive_conductance.ordinary_diffusion
        for geom in self.project.geometries().values():
            phys = GenericPhysics(network=self.network,
                                  phase=phase, geometry=geom)
            phys.add_model(propname='throat.diffusive_conductance', model=mod)
        for bcs in self.settings['inlets'].keys():
            Diff = FickianDiffusion(network=self.project.network, phase=phase)
            Pin = self.network.pores(self.settings['inlets'][bcs])
            Diff.set_value_BC(pores=Pin, values=1.0)
            Pout = self.network.pores(self.settings['outlets'][bcs])
            Diff.set_value_BC(pores=Pout, values=0.0)
            Diff.run()
            A = self.settings['areas'][bcs]
            if A is None:
                A = Diff._get_domain_area()
                self.settings['areas'][bcs] = A
            L = self.settings['lengths'][bcs]
            if L is None:
                L = Diff._get_domain_length()
                self.settings['lengths'][bcs] = L
            R = Diff.rate(pores=Pin)
            Deff = R*L/A  # Conc gradient and diffusivity were both unity
            self.results[bcs] = 1/Deff[0]

    def set_inlets(self, direction, label):
        r"""
        Specifies which pore labels indicate the inlet face for a given
        direction

        Parameters
        ----------
        direction : string
            Either 'x', 'y', or 'z', indicating the principle direction
        label : string
            The pore label indicating which pores define the inlet face

        Notes
        -----
        By default inlets are taken as pores labelled 'left', 'top' and
        'front'. If your network does not have these labels you should
        either add them, or change which labels this algorithm looks for
        using this function, or manual changing the values in the
        ``settings['inlets']``.
        """
        self.settings['inlets'][direction] = label

    def set_outlets(self, direction, label):
        r"""
        Specifies which pore labels indicate the outlet face for a given
        direction

        Parameters
        ----------
        direction : string
            Either 'x', 'y', or 'z', indicating the principle direction
        label : string
            The pore label indicating which pores define the outlet face

        Notes
        -----
        By default outlets are taken as pores labelled 'right', 'bottom' and
        'back'. If your network does not have these labels you should
        either add them, or change which labels this algorithm looks for
        using this function, or manual changing the values in the
        ``settings['outlets']``.
        """
        self.settings['outlets'][direction] = label

    def set_area(self, direction, area):
        r"""
        Specifies the area of the inlet faces normal to the given direction

        Parameters
        ----------
        direction : string
            Either 'x', 'y', or 'z', indicating the principle direction
        area : scalar
            The numerical value of the inlet face area

        Notes
        -----
        If this value is not provided then an attempt is made to calculate it
        automatically, though this estimate is not perfectly accurate,
        especially for small networks.  This value is placed into the object's
        setting dictionary under 'areas'.
        """
        self.settings['areas'][direction] = area

    def set_length(self, direction, length):
        r"""
        Specifies the length of the domain along the given direction

        Parameters
        ----------
        direction : string
            Either 'x', 'y', or 'z', indicating the principle direction
        area : scalar
            The numerical value of the domain length

        Notes
        -----
        If this value is not provided then an attempt is made to calculate it
        automatically, though this estimate is not perfectly accurate,
        especially for small networks.  This value is placed into the object's
        setting dictionary under 'lengths'.
        """
        self.settings['lengths'][direction] = length



from openpnm.network import DelaunayVoronoiDual
from openpnm import topotools
from openpnm.utils import logging
logger = logging.getLogger(__name__)


class Voronoi(DelaunayVoronoiDual):
    r"""
    Random network formed by Voronoi tessellation of arbitrary base points

    Parameters
    ----------
    points : array_like, optional
        The base points around which to generate the Voronoi tessellation.

    num_points : scalar, optional
        If ``points`` is not supplied, then this must be given.  A sent of
        randomly located points will be generated.

    shape : array_like
        The size of the domain.  It's possible to create cubic as well as 2D
        square domains by changing the ``shape`` as follows:

        [x, y, z] - will produce a normal cubic domain of dimension x, and
        and z

        [x, y, 0] - will produce a 2D square domain of size x by y

    name : string
        An optional name for the object to help identify it.  If not given,
        one will be generated.

    project : OpenPNM Project object, optional
        Each OpenPNM object must be part of a *Project*.  If none is supplied
        then one will be created and this Network will be automatically
        assigned to it.  To create a *Project* use ``openpnm.Project()``.

    Notes
    -----
    By definition these points will each lie in the center of a Voronoi cell,
    so they will not be the pore centers.  The number of pores in the
    returned network thus will differ from the number of points supplied

    """
    def __init__(self, shape=None, num_points=None, **kwargs):
        # Clean-up input points
        points = kwargs.pop('points', None)
        points = self._parse_points(shape=shape,
                                    num_points=num_points,
                                    points=points)
        # Initialize network object
        super().__init__(shape=shape, points=points, **kwargs)
        topotools.trim(network=self, pores=self.pores('delaunay'))
        pop = ['pore.delaunay', 'throat.delaunay', 'throat.interconnect']
        for item in pop:
            del self[item]



import warnings
import heapq as hq
import scipy as sp
import numpy as np
from collections import namedtuple
from openpnm.utils import logging
from openpnm.topotools import find_clusters
from openpnm.algorithms import GenericAlgorithm
logger = logging.getLogger(__name__)


class InvasionPercolation(GenericAlgorithm):
    r"""
    A classic/basic invasion percolation algorithm optimized for speed.

    Parameters
    ----------
    network : OpenPNM Network object
        The Network upon which the invasion will occur.

    Notes
    ----
    This algorithm uses a binary heap to store all a list of all accessible
    throats, sorted according to entry pressure.  This means that item [0] in
    the heap is the most easily invaded throat, so looking up which throat
    to invade next is computationally trivial.  In order to keep the list
    sorted new throats to the list takes more time, however, the heap data
    structure is very efficient at this.  Interested users can consult the
    wikipedia page on `binary heaps
    <https://en.wikipedia.org/wiki/Binary_heap>`_ for more information.


    Examples
    --------
    Start by importing the usual packages:

    >>> import openpnm as op
    >>> import scipy as sp
    >>> import matplotlib.pyplot as plt

    Create 2D cubic network for easier visualizaiton:

    >>> S = np.array([100, 100, 1])
    >>> pn = op.network.Cubic(shape=S, spacing=0.0001, name='pn11')

    Add a basic geometry:

    >>> geom = op.geometry.StickAndBall(network=pn, pores=pn.Ps, throats=pn.Ts)

    Create an invading phase, and attach the capillary pressure model:

    >>> water = op.phases.Water(network=pn)
    >>> water.add_model(propname='throat.entry_pressure',
    ...                 model=op.models.physics.capillary_pressure.washburn)

    Initialize an invasion percolation object and define inlets:

    >>> ip = op.algorithms.InvasionPercolation(network=pn)
    >>> ip.setup(phase=water)
    >>> ip.set_inlets(pores=0)
    >>> ip.run()

    After running the algorithm the invading phase configuration at a given
    saturation can be obtained and assigned to the phase object:

    >>> water.update(ip.results(Snwp=0.5))

    Because it was a 2D network it's easy to quickly visualize the invasion
    pattern as an image for verification:

    .. note::

        Because the network is 2D and cubic, an image can be generated with
        color corresponding to a value.  The following plots the entire
        invasion sequence, and the water configuraiton at Snwp = 0.5.

        ``plt.subplot(1, 2, 1)``

        ``plt.imshow(np.reshape(ip['pore.invasion_sequence'], newshape=S[S > 1]))``

        ``plt.subplot(1, 2, 2)``

        ``plt.imshow(np.reshape(water['pore.occupancy'], newshape=S[S > 1]))``

    """
    def __init__(self, settings={}, phase=None, **kwargs):
        def_set = {'phase': None,
                   'pore_volume': 'pore.volume',
                   'throat_volume': 'throat.volume',
                   'entry_pressure': 'throat.entry_pressure',
                   'gui': {'setup':          {'phase': None,
                                              'entry_pressure': '',
                                              'pore_volume': '',
                                              'throat_volume': ''},
                           'set_inlets':     {'pores': None,
                                              'overwrite': False},
                           'apply_trapping': {'outlets': None}
                           }
                   }
        super().__init__(**kwargs)
        self.settings.update(def_set)
        self.settings.update(settings)
        if phase is not None:
            self.setup(phase=phase)

    def setup(self, phase, entry_pressure='', pore_volume='', throat_volume=''):
        r"""
        Set up the required parameters for the algorithm

        Parameters
        ----------
        phase : OpenPNM Phase object
            The phase to be injected into the Network.  The Phase must have the
            capillary entry pressure values for the system.

        entry_pressure : string
            The dictionary key to the capillary entry pressure.  If none is
            supplied then the current value is retained. The default is
            'throat.capillary_pressure'.

        pore_volume : string
            The dictionary key to the pore volume.  If none is supplied then
            the current value is retained. The default is 'pore.volume'.

        throat_volume : string
            The dictionary key to the throat volume.  If none is supplied then
            the current value is retained. The default is 'throat.volume'.

        """
        self.settings['phase'] = phase.name
        if pore_volume:
            self.settings['pore_volume'] = pore_volume
        if throat_volume:
            self.settings['throat_volume'] = throat_volume
        if entry_pressure:
            self.settings['entry_pressure'] = entry_pressure

        # Setup arrays and info
        self['throat.entry_pressure'] = phase[self.settings['entry_pressure']]
        # Indices into t_entry giving a sorted list
        self['throat.sorted'] = np.argsort(self['throat.entry_pressure'], axis=0)
        self['throat.order'] = 0
        self['throat.order'][self['throat.sorted']] = np.arange(0, self.Nt)
        self['throat.invasion_sequence'] = -1
        self['pore.invasion_sequence'] = -1

    def set_inlets(self, pores=[], overwrite=False):
        r"""

        Parameters
        ----------
        pores : array_like
            The list of inlet pores from which the Phase can enter the Network
        """
        if overwrite:
            self['pore.invasion_sequence'] = -1
        self['pore.invasion_sequence'][pores] = 0

        # Perform initial analysis on input pores
        Ts = self.project.network.find_neighbor_throats(pores=pores)
        self.queue = []
        [hq.heappush(self.queue, T) for T in self['throat.order'][Ts]]

    def run(self, n_steps=None):
        r"""
        Perform the algorithm

        Parameters
        ----------
        n_steps : int
            The number of throats to invaded during this step

        """
        if n_steps is None:
            n_steps = sp.inf

        if len(self.queue) == 0:
            logger.warn('queue is empty, this network is fully invaded')
            return

        # Create incidence matrix to get neighbor throats later in _run method
        incidence_matrix = self.network.create_incidence_matrix(fmt='csr')
        t_inv, p_inv, p_inv_t = InvasionPercolation._run_accelerated(
            queue=self.queue,
            t_sorted=self['throat.sorted'],
            t_order=self['throat.order'],
            t_inv=self['throat.invasion_sequence'],
            p_inv=self['pore.invasion_sequence'],
            p_inv_t=np.zeros_like(self['pore.invasion_sequence']),
            conns=self.project.network['throat.conns'],
            idx=incidence_matrix.indices,
            indptr=incidence_matrix.indptr,
            n_steps=n_steps
        )

        self['throat.invasion_sequence'] = t_inv
        self['pore.invasion_sequence'] = p_inv
        self['throat.invasion_pressure'] = self['throat.entry_pressure']
        self['pore.invasion_pressure'] = self['throat.entry_pressure'][p_inv_t]
        self['pore.invasion_pressure'][self['pore.invasion_sequence']==0] = 0.0

    def results(self, Snwp=None):
        r"""
        Returns the phase configuration at the specified non-wetting phase
        (invading phase) saturation.

        Parameters
        ----------
        Snwp : scalar, between 0 and 1
            The network saturation for which the phase configuration is
            desired.

        Returns
        -------
        Two dictionary containing arrays that describe the pore and throat
        distribution at the given saturation.  Specifically, these are:

        **'pore.occupancy'** : 1 indicates the pores is invaded and 0
        otherwise.

        **'throat.occupancy'** : Same as described above but for throats.

        """
        if Snwp is None:
            Np = self['pore.invasion_sequence']
            Nt = self['throat.invasion_sequence']
            data = {'pore.invasion_sequence': Np,
                    'throat.invasion_sequence': Nt}
        else:
            net = self.project.network
            P12 = net['throat.conns']
            # Fetch void volume for pores and throats
            Vp = net[self.settings['pore_volume']]
            Vt = net[self.settings['throat_volume']]
            # Fetch the order of filling
            Np = self['pore.invasion_sequence']
            Nt = self['throat.invasion_sequence']
            # Create Nt-long mask of which pores were filled when throat was filled
            Pinv = (Np[P12].T == Nt).T
            # If a pore and throat filled together, find combined volume
            Vinv = np.vstack(((Pinv*Vp[P12]).T, Vt)).T
            Vinv = np.sum(Vinv, axis=1)
            # Convert to cumulative volume filled as each throat is invaded
            x = np.argsort(Nt)  # Find order throats were invaded
            Vinv_cum = np.cumsum(Vinv[x])
            # Normalized cumulative volume filled into saturation
            S = Vinv_cum/(Vp.sum() + Vt.sum())
            # Find throat invasion step where Snwp was reached
            try:
                N = np.where(S < Snwp)[0][-1]
            except Exception:
                N = -np.inf
            data = {'pore.occupancy': Np <= N, 'throat.occupancy': Nt <= N}
        return data

    def apply_trapping(self, outlets):
        """
        Apply trapping based on algorithm described by Y. Masson [1].
        It is applied as a post-process and runs the percolation algorithm in
        reverse assessing the occupancy of pore neighbors. Consider the
        following scenario when running standard IP without trapping,
        3 situations can happen after each invasion step:

            * The number of defending clusters stays the same and clusters can
            shrink
            * A cluster of size one is suppressed
            * A cluster is split into multiple clusters

        In reverse the following opposite situations can happen:

            * The number of defending clusters stays the same and clusters can
            grow
            * A cluster of size one is created
            * Mutliple clusters merge into one cluster

        With trapping the reversed rules are adjusted so that only clusters
        that do not connect to a sink can grow and merge. At the point that a
        neighbor connected to a sink is touched the trapped cluster stops
        growing as this is the point of trapping in forward invasion time.

        Logger info displays the invasion sequence and pore index and a message
        with condition number based on the modified trapping rules and the
        assignment of the pore to a given cluster.

        Initially all invaded pores are given cluster label -1
        Outlets / Sinks are given -2
        New clusters that grow into fully trapped clusters are either
        identified at the point of breakthrough or grow from nothing if the
        full invasion sequence is run, they are assigned numbers from 0 up.

        Ref:
        [1] Masson, Y., 2016. A fast two-step algorithm for invasion
        percolation with trapping. Computers & Geosciences, 90, pp.41-48

        Parameters
        ----------
        outlets : list or array of pore indices for defending fluid to escape
        through

        Returns
        -------
        Creates a throat array called 'pore.clusters' in the Algorithm
        dictionary. Any positive number is a trapped cluster

        Also creates 2 boolean arrays Np and Nt long called '<element>.trapped'
        """
        # First see if network is fully invaded
        net = self.project.network
        invaded_ps = self['pore.invasion_sequence'] > -1
        if ~np.all(invaded_ps):
            # Put defending phase into clusters
            clusters = find_clusters(network=net, mask=~invaded_ps)
            # Identify clusters that are connected to an outlet and set to -2
            # -1 is the invaded fluid
            # -2 is the defender fluid able to escape
            # All others now trapped clusters which grow as invasion is reversed
            out_clusters = np.unique(clusters[outlets])
            for c in out_clusters:
                if c >= 0:
                    clusters[clusters == c] = -2
        else:
            # Go from end
            clusters = np.ones(net.Np, dtype=int)*-1
            clusters[outlets] = -2

        # Turn into a list for indexing
        inv_seq = np.vstack((self['pore.invasion_sequence'].astype(int),
                             np.arange(0, net.Np, dtype=int))).T
        # Reverse sort list
        inv_seq = inv_seq[inv_seq[:, 0].argsort()][::-1]
        next_cluster_num = np.max(clusters)+1
        # For all the steps after the inlets are set up to break-through
        # Reverse the sequence and assess the neighbors cluster state
        stopped_clusters = np.zeros(net.Np, dtype=bool)
        all_neighbors = net.find_neighbor_pores(net.pores(), flatten=False,
                                                include_input=True)
        for un_seq, pore in inv_seq:
            if pore not in outlets and un_seq > 0:  # Skip inlets and outlets
                nc = clusters[all_neighbors[pore]]  # Neighboring clusters
                unique_ns = np.unique(nc[nc != -1])  # Unique Neighbors
                seq_pore = "S:"+str(un_seq)+" P:"+str(pore)
                if np.all(nc == -1):
                    # This is the start of a new trapped cluster
                    clusters[pore] = next_cluster_num
                    next_cluster_num += 1
                    msg = (seq_pore+" C:1 new cluster number: "
                           + str(clusters[pore]))
                    logger.info(msg)
                elif len(unique_ns) == 1:
                    # Grow the only connected neighboring cluster
                    if not stopped_clusters[unique_ns[0]]:
                        clusters[pore] = unique_ns[0]
                        msg = (seq_pore+" C:2 joins cluster number: "
                               + str(clusters[pore]))
                        logger.info(msg)
                    else:
                        clusters[pore] = -2
                elif -2 in unique_ns:
                    # We have reached a sink neighbor, stop growing cluster
                    msg = (seq_pore+" C:3 joins sink cluster")
                    logger.info(msg)
                    clusters[pore] = -2
                    # Stop growth and merging
                    stopped_clusters[unique_ns[unique_ns > -1]] = True
                else:
                    # We might be able to do some merging
                    # Check if any stopped clusters are neighbors
                    if np.any(stopped_clusters[unique_ns]):
                        msg = (seq_pore+" C:4 joins sink cluster")
                        logger.info(msg)
                        clusters[pore] = -2
                        # Stop growing all neighboring clusters
                        stopped_clusters[unique_ns] = True
                    else:
                        # Merge multiple un-stopped trapped clusters
                        new_num = unique_ns[0]
                        clusters[pore] = new_num
                        for c in unique_ns:
                            clusters[clusters == c] = new_num
                            msg = (seq_pore + " C:5 merge clusters: "
                                   + str(c) + " into "+str(new_num))
                            logger.info(msg)

        # And now return clusters
        self['pore.clusters'] = clusters
        logger.info("Number of trapped clusters"
                    + str(np.sum(np.unique(clusters) >= 0)))
        self['pore.trapped'] = self['pore.clusters'] > -1
        trapped_ts = net.find_neighbor_throats(self['pore.trapped'])
        self['throat.trapped'] = np.zeros([net.Nt], dtype=bool)
        self['throat.trapped'][trapped_ts] = True
        self['pore.invasion_sequence'][self['pore.trapped']] = -1
        self['throat.invasion_sequence'][self['throat.trapped']] = -1

    def get_intrusion_data(self):
        r"""
        Get the percolation data as the invader volume or number fraction vs
        the capillary capillary pressure.

        """
        if 'pore.invasion_pressure' not in self.props():
            logger.error('Algorithm must be run first')
            return None
        net = self.project.network
        pvols = net[self.settings['pore_volume']]
        tvols = net[self.settings['throat_volume']]
        tot_vol = np.sum(pvols) + np.sum(tvols)
        # Normalize
        pvols /= tot_vol
        tvols /= tot_vol
        # Remove trapped volume
        pvols[self['pore.invasion_sequence'] == -1] = 0.0
        tvols[self['throat.invasion_sequence'] == -1] = 0.0
        pseq = self['pore.invasion_sequence']
        tseq = self['throat.invasion_sequence']
        tPc = self['throat.invasion_pressure']
        pPc = self['pore.invasion_pressure']
        # Change the entry pressure for trapped pores and throats to be 0
        pPc[self['pore.invasion_sequence'] == -1] = 0.0
        tPc[self['throat.invasion_sequence'] == -1] = 0.0
        vols = np.concatenate((pvols, tvols))
        seqs = np.concatenate((pseq, tseq))
        Pcs = np.concatenate((pPc, tPc))
        data = np.rec.fromarrays([seqs, vols, Pcs], formats=['i', 'f', 'f'],
                                 names=['seq', 'vol', 'Pc'])
        data.sort(axis=0, order='seq')
        sat = np.cumsum(data.vol)
        pc_curve = namedtuple('pc_curve', ('Pcap', 'S_tot'))
        data = pc_curve(data.Pc, sat)
        return data

    def plot_intrusion_curve(self, fig=None):
        r"""
        Plot the percolation curve as the invader volume or number fraction vs
        the capillary capillary pressure.

        """
        import matplotlib.pyplot as plt

        data = self.get_intrusion_data()
        if data is not None:
            if fig is None:
                fig, ax = plt.subplots()
            else:
                ax = fig.gca()
            ax.semilogx(data.Pcap, data.S_tot,)
            plt.ylabel('Invading Phase Saturation')
            plt.xlabel('Capillary Pressure')
            plt.grid(True)
            return fig
        else:
            return None

    def _run_accelerated(queue, t_sorted, t_order, t_inv, p_inv, p_inv_t,
                         conns, idx, indptr, n_steps):
        r"""
        Numba-jitted run method for InvasionPercolation class.

        Notes
        -----
        (1) ``idx`` and ``indptr`` are properties are the network's incidence
        matrix, and are used to quickly find neighbor throats.

        (2) Numba doesn't like forein data types (i.e. GenericNetwork), and so
        ``find_neighbor_throats`` method cannot be called in a jitted method.

        (3) Nested wrapper is for performance issues (reduced OpenPNM import)
        time due to local numba import

        """
        from numba import njit
        try:
            from numba.core.errors import NumbaPendingDeprecationWarning
        except ModuleNotFoundError:
            from numba.errors import NumbaPendingDeprecationWarning
        warnings.simplefilter('ignore', category=NumbaPendingDeprecationWarning)

        @njit
        def wrapper(queue, t_sorted, t_order, t_inv, p_inv, p_inv_t, conns,
                    idx, indptr, n_steps):
            count = 0
            while (len(queue) > 0) and (count < n_steps):
                # Find throat at the top of the queue
                t = hq.heappop(queue)
                # Extract actual throat number
                t_next = t_sorted[t]
                t_inv[t_next] = count
                # If throat is duplicated
                while len(queue) > 0 and queue[0] == t:
                    # Note: Preventing duplicate entries below might save some time
                    t = hq.heappop(queue)
                # Find pores connected to newly invaded throat
                Ps = conns[t_next]
                # Remove already invaded pores from Ps
                Ps = Ps[p_inv[Ps] < 0]
                if len(Ps) > 0:
                    p_inv[Ps] = count
                    p_inv_t[Ps] = t_next
                    for i in Ps:
                        Ts = idx[indptr[i]:indptr[i+1]]
                        Ts = Ts[t_inv[Ts] < 0]
                    for i in set(Ts):   # set(Ts) to exclude repeated neighbor throats
                        hq.heappush(queue, t_order[i])
                count += 1
            return t_inv, p_inv, p_inv_t

        return wrapper(queue, t_sorted, t_order, t_inv, p_inv, p_inv_t, conns,
                       idx, indptr, n_steps)


if __name__ == '__main__':
    import openpnm as op
    pn = op.network.Cubic(shape=[10, 10, 10], spacing=1e-4)
    geo = op.geometry.StickAndBall(network=pn, pores=pn.Ps, throats=pn.Ts)
    water = op.phases.Water(network=pn, name='h2o')
    phys_water = op.physics.Standard(network=pn, phase=water, geometry=geo)
    ip = InvasionPercolation(network=pn, phase=water)
    ip.set_inlets(pn.pores('left'))
    ip.run()
    ip.plot_intrusion_curve()



from openpnm.algorithms import OrdinaryPercolation
from openpnm.utils import logging
import numpy as np
logger = logging.getLogger(__name__)


class Porosimetry(OrdinaryPercolation):
    r"""
    Simulates mercury instrustion porosimetry using ordinary percolation

    Parameters
    ----------
    network : OpenPNM Network object
        The Network upon which this simulation should be run

    name : string, optional
        An identifying name for the object.  If none is given then one is
        generated.

    project : OpenPNM Project object
        Either a Network or a Project must be specified

    Notes
    -----
    Mercury intrusion progresses by applying increasing pressures to the
    invading mercury phase, and measuring the resultant volume of invading
    fluid.  This corresponds directly to an ordinary percolation process,
    with access limitations enabled.

    See Also
    --------
    OrdinaryPercolation


    """

    def __init__(self, settings={}, phase=None, **kwargs):
        def_set = {'phase': None,
                   'pore_volume': 'pore.volume',
                   'throat_volume': 'throat.volume',
                   'mode': 'bond',
                   'access_limited': True,
                   'quantity': 'pressure',
                   'throat_entry_pressure': 'throat.entry_pressure',
                   'pore_volume': 'pore.volume',
                   'throat_volume': 'throat.volume',
                   'late_pore_filling': '',
                   'late_throat_filling': '',
                   'gui': {'setup': {'phase': None,
                                     'quantity': '',
                                     'throat_entry_pressure': '',
                                     'pore_volume': '',
                                     'throat_volume': '',
                                     'late_pore_filling': '',
                                     'late_throat_filling': ''},
                           'set_inlets':   {'pores': None,
                                            'overwrite': False},
                           'set_outlets':  {'pores': None,
                                            'overwrite': False},
                           'set_residual': {'pores': None,
                                            'throats': None,
                                            'overwrite': False}
                           }
                   }
        super().__init__(**kwargs)
        self.settings.update(def_set)
        # Apply user settings, if any
        self.settings.update(settings)
        # Use the reset method to initialize all arrays
        self.reset()
        if phase is not None:
            self.setup(phase=phase)

    def setup(self,
              phase=None,
              quantity='',
              throat_entry_pressure='',
              pore_volume='',
              throat_volume='',
              late_pore_filling='',
              late_throat_filling=''):
        r"""
        Used to specify necessary arguments to the simulation.  This method is
        useful for resetting the algorithm or applying more explicit control.

        Parameters
        ----------
        phase : OpenPNM Phase object
            The Phase object containing the physical properties of the invading
            fluid.

        quantity : string
            The name of the quantity calculated by this algorithm.  This is
            used for instance, by the late pore and throat filling models
            to indicate the prevailing fluid pressure in the invading phase
            for calculating the extent of filling.  The default is
            'pressure'.  Note that there is no need to specify 'pore' and/or
            'throat' with this as the given value will apply to both.

        throat_entry_pressure : string
            The dictionary key on the Phase object where the throat entry
            pressure values are stored.  The default is
            'throat.entry_pressure'.

        pore_volume : string
            The dictionary key containing the pore volume information. The
            default is 'pore.volume'.

        throat_volume : string
            The dictionary key containing the throat volume information.  The
            default is 'throat.volume'.

        pore_partial_filling : string
            The name of the model used to determine partial pore filling as
            a function of applied pressure.

        throat_partial_filling : string
            The name of the model used to determine partial throat filling as
            a function of applied pressure.

        """
        if phase:
            self.settings['phase'] = phase.name
        if quantity:
            self.settings['quantity'] = quantity
        if throat_entry_pressure:
            self.settings['throat_entry_pressure'] = throat_entry_pressure
            phase = self.project.find_phase(self)
            self['throat.entry_pressure'] = phase[throat_entry_pressure]
        if pore_volume:
            self.settings['pore_volume'] = pore_volume
        if throat_volume:
            self.settings['throat_volume'] = throat_volume
        if late_pore_filling:
            self.settings['late_pore_filling'] = late_pore_filling
        if late_throat_filling:
            self.settings['late_throat_filling'] = late_throat_filling

    def set_partial_filling(self, propname):
        r"""
        Define which pore filling model to apply.

        Parameters
        ----------
        propname : string
            Dictionary key on the physics object(s) containing the pore
            filling model(s) to apply.

        Notes
        -----
        It is assumed that these models are functions of the `quantity`
        specified in the algorithms settings.  This values is applied to the
        corresponding phase just prior to regenerating the given pore-scale
        model(s).

        """
        if propname.startswith('pore'):
            self.settings['pore_partial_filling'] = propname
        if propname.startswith('throat'):
            self.settings['throat_partial_filling'] = propname

    def run(self, points=25, start=None, stop=None):
        if self.settings['mode'] != 'bond':
            raise Exception('Porosimetry must be run as bond percolation')
        if self.settings['access_limited'] is False:
            raise Exception('Porosimetry must be run as access limited')
        super().run(points=points, start=start, stop=stop)

    run.__doc__ = OrdinaryPercolation.run.__doc__

    def results(self, Pc=None):
        r"""
        """
        if Pc is None:
            p_inv = self['pore.invasion_pressure']
            t_inv = self['throat.invasion_pressure']
            results = {'pore.invasion_pressure': p_inv,
                       'throat.invasion_pressure': t_inv}
        else:
            p_inv, t_inv = super().results(Pc).values()
            phase = self.project.find_phase(self)
            quantity = self.settings['quantity'].split('.')[-1]
            lpf = np.array([1])
            if self.settings['pore_partial_filling']:
                # Set pressure on phase to current capillary pressure
                phase['pore.'+quantity] = Pc
                # Regenerate corresponding physics model
                for phys in self.project.find_physics(phase=phase):
                    phys.regenerate_models(self.settings['pore_partial_filling'])
                # Fetch partial filling fraction from phase object (0->1)
                lpf = phase[self.settings['pore_partial_filling']]
            # Calculate filled throat volumes
            ltf = np.array([1])
            if self.settings['throat_partial_filling']:
                # Set pressure on phase to current capillary pressure
                phase['throat.'+quantity] = Pc
                # Regenerate corresponding physics model
                for phys in self.project.find_physics(phase=phase):
                    phys.regenerate_models(self.settings['throat_partial_filling'])
                # Fetch partial filling fraction from phase object (0->1)
                ltf = phase[self.settings['throat_partial_filling']]
            p_inv = p_inv*lpf
            t_inv = t_inv*ltf
            results = {'pore.occupancy': p_inv, 'throat.occupancy': t_inv}
        return results



import numpy as np
import scipy as sp
from openpnm.algorithms import ReactiveTransport
from openpnm.utils import logging, GenericSettings, Docorator
logger = logging.getLogger(__name__)
docstr = Docorator()


@docstr.get_sectionsf('StokesFlowSettings',
                      sections=['Parameters'])
@docstr.dedent
class StokesFlowSettings(GenericSettings):
    r"""

    Parameters
    ----------
    %(GenericTransportSettings.parameters)s
    quantity : str (default = 'pore.pressure')
        The name of the physical quantity to be calculated
    conductance : str (default = 'throat.hydraulic_conductance')
        The name of the pore-scale transport conductance values. These are
        typically calculated by a model attached to a *Physics* object
        associated with the given *Phase*.

    Other Parameters
    ----------------

    **The following parameters pertain to the ReactiveTransport class**

    %(ReactiveTransportSettings.other_parameters)s

    ----

    **The following parameters pertain to the GenericTransport class**

    %(GenericTransportSettings.other_parameters)s

    """
    quantity = 'pore.pressure'
    conductance = 'throat.hydraulic_conductance'


class StokesFlow(ReactiveTransport):
    r"""
    A subclass of GenericLinearTransport to simulate viscous flow.

    """

    def __init__(self, settings={}, **kwargs):
        super().__init__(**kwargs)
        self.settings._update_settings_and_docs(StokesFlowSettings())
        self.settings.update(settings)

    def setup(self, phase=None, quantity='', conductance='', **kwargs):
        r"""
        This method takes several arguments that are essential to running the
        algorithm and adds them to the settings.

        Parameters
        ----------
        %(StokesFlowSettings.parameters)s

        Notes
        -----
        Any additional arguments are added to the ``settings`` dictionary of
        the object.

        """
        if phase:
            self.settings['phase'] = phase.name
        if quantity:
            self.settings['quantity'] = quantity
        if conductance:
            self.settings['conductance'] = conductance
        super().setup(**kwargs)

    def calc_effective_permeability(self, inlets=None, outlets=None,
                                    domain_area=None, domain_length=None):
        r"""
        This calculates the effective permeability in this linear transport
        algorithm.

        Parameters
        ----------
        inlets : array_like
            The pores where the inlet pressure boundary conditions were
            applied.  If not given an attempt is made to infer them from the
            algorithm.

        outlets : array_like
            The pores where the outlet pressure boundary conditions were
            applied.  If not given an attempt is made to infer them from the
            algorithm.

        domain_area : scalar, optional
            The area of the inlet (and outlet) boundary faces.  If not given
            then an attempt is made to estimate it, but it is usually
            underestimated.

        domain_length : scalar, optional
            The length of the domain between the inlet and outlet boundary
            faces.  If not given then an attempt is made to estimate it, but it
            is usually underestimated.

        Notes
        -----
        The area and length of the domain are found using the bounding box
        around the inlet and outlet pores which do not necessarily lie on the
        edge of the domain, resulting in underestimation of sizes.
        """
        phase = self.project.phases()[self.settings['phase']]
        d_normal = self._calc_eff_prop(inlets=inlets, outlets=outlets,
                                       domain_area=domain_area,
                                       domain_length=domain_length)
        K = d_normal * np.mean(phase['pore.viscosity'])
        return K



import numpy as np
import openpnm.models.misc as misc
from openpnm.phases import GenericPhase as GenericPhase
from openpnm.utils import logging
logger = logging.getLogger(__name__)


class MultiPhase(GenericPhase):
    r"""
    Creates Phase object that represents a multiphase system consisting of
    a given list of OpenPNM Phase objects.

    Parameters
    ----------
    network : OpenPNM Network object
        The network to which this phase object will be attached.

    project : OpenPNM Project object, optional
        The Project with which this phase should be associted.  If a
        ``network`` is given then this is ignored and the Network's project
        is used.  If a ``network`` is not given then this is mandatory.

    name : string, optional
        The name of the phase.  This is useful to keep track of the objects
        throughout the simulation.  The name must be unique to the project.
        If no name is given, one is generated.

    Examples
    --------
    >>> import scipy as sp
    >>> import openpnm as op
    >>> pn = op.network.Cubic(shape=[5, 5, 5])
    >>> air = op.phases.Air(network=pn, name='air')  # Create two pure phases
    >>> water = op.phases.Water(network=pn, name='water')
    >>> m = op.phases.MultiPhase(network=pn, phases=[air, water], name='multi')
    >>> Ps = pn['pore.coords'][:, 0] < 3  # Pick some pores to be air filled
    >>> Ts = pn.find_neighbor_throats(pores=Ps)  # Find neighboring throats
    >>> Ts = pn.tomask(throats=Ts)  # Convert throat indices to mask
    >>> m.set_occupancy(phase=air, Pvals=Ps, Tvals=Ts)  # Assign occupancies
    >>> m.set_occupancy(phase=water, Pvals=~Ps, Tvals=~Ts)

    Air and water have uniform viscosity values throughout, so the peak-to-peak
    distance is 0, while the mixture phase has the viscosity of air in some
    locations and water in others, hence a heterogenous viscosity array:

    >>> np.ptp(water['pore.viscosity']) == 0
    True
    >>> np.ptp(air['pore.viscosity']) == 0
    True
    >>> np.ptp(m['pore.viscosity']) == 0
    False

    """
    def __init__(self, phases=[], settings={}, **kwargs):
        super().__init__(**kwargs)
        self.settings.update({'phases': [],
                              'throat_occupancy': 'manual',
                              'partition_coef': 'throat.partition_coef.all'})
        self.settings.update(settings)

        self['pore.occupancy.all'] = np.zeros(self.Np, dtype=float)
        self['throat.occupancy.all'] = np.zeros(self.Nt, dtype=float)

        self.pop('pore.temperature', None)
        self.pop('pore.pressure', None)
        # Initialize throat partition coefficient values to 1.0
        partition_coef = self.settings['partition_coef']
        self[partition_coef] = np.ones(self.Nt)

        # Add supplied phases to the phases list and initialize occupancy to 0
        self.add_phases(phases)

        logger.warning('MultiPhase is a beta feature. Functionality may change!')

    def __getitem__(self, key):
        try:
            vals = super().__getitem__(key)
        except KeyError:
            vals = self.interleave_data(key)
        return vals

    def _update_occupancy(self):
        # Update "occupancy.all" by summing up all occupancies
        for elem in ["pore", "throat"]:
            dict_ = self[f"{elem}.occupancy"]
            dict_.pop(f"{elem}.occupancy.all")
            self[f"{elem}.occupancy.all"] = np.sum(list(dict_.values()), axis=0)

    def _get_phases(self):
        phases = {self.project[item].name: self.project[item]
                  for item in self.settings['phases']}
        return phases

    phases = property(fget=_get_phases)

    def add_phases(self, phases):
        r"""
        Add supplied phases to the MultiPhase object and initialize occupancy to 0.
        """
        phases = np.array(phases, ndmin=1)
        for phase in phases:
            if phase.name not in self.settings['phases']:
                self.settings['phases'].append(phase.name)
                self[f'pore.occupancy.{phase.name}'] = 0.0
                self[f'throat.occupancy.{phase.name}'] = 0.0

    def set_binary_partition_coef(self, phases, model,
                                  propname="throat.partition_coef", **kwargs):
        r"""
        Set binary partition coefficient that defines the concentration ratio
        of phase 1 to phase 2.
        """
        # Raise error if phases is not exactly two-element long
        if np.size(phases) != 2:
            raise Exception("'phases' must contain exactly two elements!")
        # Add interface model to the MultiPhase object
        self._add_interface_prop(propname, phases, model, **kwargs)
        # Get binary partition coefficient
        interface_prop_key = propname + f".{phases[0].name}:{phases[1].name}"
        K12 = self[interface_prop_key]
        # Modify the global partition coefficient vector
        network = self.project.network
        cn = network["throat.conns"]
        occ_phase1 = self[f"pore.occupancy.{phases[0].name}"][cn]
        occ_phase2 = self[f"pore.occupancy.{phases[1].name}"][cn]
        partition_coef = self.settings["partition_coef"]
        K_global = self[partition_coef]
        mask = (occ_phase1[:, 0] + occ_phase2[:, 1]) == 2.0
        K_global[mask] = K12[mask]
        mask = (occ_phase2[:, 0] + occ_phase1[:, 1]) == 2.0
        K_global[mask] = 1.0 / K12[mask]

    def _add_interface_prop(self, propname, phases, model, **kwargs):
        r"""
        Helper method used to add interface models to the MultiPhase object by
        augmenting ``propname`` --> ``propname.phases[0].name:phases[1].name``
        """
        # Add "throat" keyword to the begining of propname if no identifier is found
        if propname.split(".")[0] not in ["pore", "throat"]:
            propname = f"throat.{propname}"
        # Check propname is throat property
        if not propname.startswith("throat"):
            raise Exception("'propname' must be a throat property such as: "
                            + "'throat.partition_coef'.")
        propname += f".{phases[0].name}:{phases[1].name}"
        self.add_model(propname, model, **kwargs)

    def interleave_data(self, prop):
        r"""
        Gathers property values from component phases to build a single array

        If the requested ``prop`` is not on this MultiPhase, then a search is
        conducted on all associated phase objects, and values from each
        are assembled into a single array.

        Parameters
        ----------
        prop : string
            The property to be retrieved

        Returns
        -------
        array : ND-array
            An array containing the specified property retrieved from each
            component phase and assembled based on the specified mixing rule

        """
        element = prop.split('.')[0]
        vals = np.zeros([self._count(element=element)], dtype=float)
        # Retrieve property from constituent phases (weight = occupancy)
        try:
            for phase in self.phases.values():
                vals += phase[prop] * self[element + '.occupancy.' + phase.name]
        # Otherwise - if not found - retrieve from super class
        except KeyError:
            vals = super().interleave_data(prop)

        # Check for consistency of occypancy values (i.e. add up to 1)
        if np.any(self[element + '.occupancy.all'] != 1.0):
            self._update_occupancy()
            if np.any(self[element + '.occupancy.all'] != 1.0):
                raise Exception(f"Occupancy doesn't add to unity in all {element}s")
        return vals

    def regenerate_models(self, **kwargs):
        r"""
        Regenerate models associated with the Multiphase object

        This method works by first regenerating the models associated with the
        constituent phases, and then regenerating Multiphase models.

        """
        # Regenerate models associated with phases within MultiPhase object
        for phase in self.phases.values():
            phase.regenerate_models(**kwargs)
        # Regenerate models specific to MultiPhase object
        super().regenerate_models(**kwargs)

    def _set_automatic_throat_occupancy(self, mode="mean"):
        r"""
        Automatically interpolates throat occupancy based on adjacent pores.
        """
        self.settings['throat_occupancy'] = 'automatic'
        for phase in self.phases.values():
            self.add_model(propname=f"throat.occupancy.{phase.name}",
                           model=misc.from_neighbor_pores,
                           prop=f"pore.occupancy.{phase.name}",
                           mode=mode)

    def set_occupancy(self, phase, Pvals=[], Tvals=[], pores=[], throats=[]):
        r"""
        Specify occupancy of a phase in each pore and/or throat

        Parameters
        ----------
        phase : OpenPNM Phase object
            The phase whose occupancy is being specified

        Pvals : array_like, float
            The volume fraction of ``phase`` in each pore. This array must be
            *``Np``*-long, except when ``pores`` is also passed, where in that
            case they must be of equal length, with one value between 0 and 1,
            for each pore in the network. If a scalar is received it is applied
            to all pores. If nothing is passed, ``Pvals=1.0`` is assumed.

        Tvals : array_like, float
            The volume fraction of ``phase`` in each throat. This array must be
            *``Nt``*-long, except when ``throats`` is also passed, where in that
            case they must be of equal length, with one value between 0 and 1,
            for each throat in the network. If a scalar is received it is applied
            to all throats. If nothing is passed, ``Tvals=1.0`` is assumed.

        pores : array_like, int
            The location of pores for which the phase occupancy is to be set.

        throats : array_like, int
            The location of throats for which the phase occupancy is to be set.

        """
        # TODO: pores/throats could also be masks

        Pvals = np.array(Pvals, ndmin=1)
        Tvals = np.array(Tvals, ndmin=1)
        pores = np.array(pores, ndmin=1)
        throats = np.array(throats, ndmin=1)

        # Check for size consistency of the arguments
        if Pvals.size and pores.size:
            if Pvals.size != 1 and pores.size != 1:
                if Pvals.size != pores.size:
                    raise Exception("Pvals and pores must be the same size.")
        if Tvals.size and throats.size:
            if Tvals.size != 1 and throats.size != 1:
                if Tvals.size != throats.size:
                    raise Exception("Tvals and throats must be the same size.")

        # Check if the passed phase is already part of MultiPhase object
        if phase not in self.project:
            raise Exception(f"{phase.name} doesn't belong to this project")
        # Add the passed phase to MultiPhase object if not found
        else:
            if phase.name not in self.settings['phases']:
                self.add_phases(phase)

        # Check for value consistency of the arguments
        if np.any(Pvals > 1.0) or np.any(Pvals < 0.0):
            logger.critical('Received Pvals contain volume fractions outside '
                            + 'the range of 0 to 1')
        if np.any(Tvals > 1.0) or np.any(Tvals < 0.0):
            logger.critical('Received Tvals contain volume fractions outside '
                            + 'the range of 0 to 1')

        if Pvals.size and not pores.size:
            pores = self.pores()
        if Tvals.size and not throats.size:
            throats = self.throats()

        if pores.size:
            Pvals = Pvals if Pvals.size else 1.0
            self['pore.occupancy.' + phase.name][pores] = Pvals
        if throats.size:
            Tvals = Tvals if Tvals.size else 1.0
            self['throat.occupancy.' + phase.name][throats] = Tvals

        if self.settings["throat_occupancy"] == "automatic":
            self.regenerate_models(propnames=f"throat.occupancy.{phase.name}")
        self._update_occupancy()



# -*- coding: utf-8 -*-
"""
===============================================================================
Cubic: Generate lattice-like networks
===============================================================================

"""
import numpy as np
import scipy as sp
from openpnm.network import GenericNetwork
from openpnm import topotools
from openpnm.utils import logging

logger = logging.getLogger(__name__)


class Cubic(GenericNetwork):
    r"""
    Simple cubic lattice with connectivity from 6 to 26

    Though simple, the Cubic network offers many advantages such as easy
    visualization and accurate determination of domain area and length in
    transport calculations.

    Parameters
    ----------
    shape : array_like
        The [Nx, Ny, Nz] size of the network in terms of the number of pores in
        each direction

    spacing : array_like, optional
        The spacing between pore centers in each direction. If not given, then
        [1, 1, 1] is assumed.

    connectivity : int, optional
        The number of connections to neighboring pores.  Connections are made
        symmetrically to any combination of face, edge, or corners neighbors.
        The default is 6 to create a simple cubic structure, but options are:

        - 6: Faces only
        - 8: Corners only
        - 12: Edges only
        - 14: Faces and Corners
        - 18: Faces and Edges
        - 20: Edges and Corners
        - 26: Faces, Edges and Corners

        For a more random distribution of connectivity, use a high
        ``connectivity`` (i.e. 26) and then delete a fraction of the throats
        using ``openpnm.topotools.reduce_coordination``.

    name : string
        An optional name for the object to help identify it.  If not given,
        one will be generated.

    project : OpenPNM Project object, optional
        Each OpenPNM object must be part of a *Project*.  If none is supplied
        then one will be created and this Network will be automatically
        assigned to it.  To create a *Project* use ``openpnm.Project()``.

    Attributes
    ----------
    spacing : int or array
        The distance between pore centers.  This value becomes meaningless
        if the topology is manipulated at all (i.e. by adding boundary pores)
        since there is not unique or consistent value.  In such cases an
        exception is thrown.

    shape : array
        The shape of the network.  Like ``spacing`` this values is meaningless
        if the topology is manipulated, so an Exception is thrown.

    Examples
    --------
    >>> import openpnm as op
    >>> pn = op.network.Cubic(shape=[5, 5, 5], spacing=[1, 1, 1])
    >>> pn.Np
    125

    And it can be plotted for quick visualization using:

    >>> fig = op.topotools.plot_connections(network=pn)
    >>> fig = op.topotools.plot_coordinates(network=pn, c='r', s=75, fig=fig)

    .. image:: /../docs/static/images/cubic_network.png
        :align: center

    For larger networks and more control over presentation use `Paraview
    <http://www.paraview.org>`_.
    """

    def __init__(
        self, shape, spacing=[1, 1, 1], connectivity=6, name=None, project=None
    ):

        # Take care of 1D/2D networks
        shape = np.array(shape, ndmin=1)
        shape = np.concatenate((shape, [1] * (3 - shape.size))).astype(int)

        arr = np.atleast_3d(np.empty(shape))

        # Store original network shape
        self._shape = np.shape(arr)
        # Store network spacing
        spacing = sp.float64(spacing)
        if spacing.size == 2:
            spacing = np.concatenate((spacing, [1]))
        self._spacing = np.ones(3, dtype=float) * np.array(spacing, ndmin=1)

        z = np.tile(np.arange(shape[2]), shape[0] * shape[1])
        y = np.tile(np.repeat(np.arange(shape[1]), shape[2]), shape[0])
        x = np.repeat(np.arange(shape[0]), shape[1] * shape[2])
        points = (np.vstack([x, y, z]).T).astype(float) + 0.5

        idx = np.arange(arr.size).reshape(arr.shape)

        face_joints = [
            (idx[:, :, :-1], idx[:, :, 1:]),
            (idx[:, :-1], idx[:, 1:]),
            (idx[:-1], idx[1:]),
        ]

        corner_joints = [
            (idx[:-1, :-1, :-1], idx[1:, 1:, 1:]),
            (idx[:-1, :-1, 1:], idx[1:, 1:, :-1]),
            (idx[:-1, 1:, :-1], idx[1:, :-1, 1:]),
            (idx[1:, :-1, :-1], idx[:-1, 1:, 1:]),
        ]

        edge_joints = [
            (idx[:, :-1, :-1], idx[:, 1:, 1:]),
            (idx[:, :-1, 1:], idx[:, 1:, :-1]),
            (idx[:-1, :, :-1], idx[1:, :, 1:]),
            (idx[1:, :, :-1], idx[:-1, :, 1:]),
            (idx[1:, 1:, :], idx[:-1, :-1, :]),
            (idx[1:, :-1, :], idx[:-1, 1:, :]),
        ]

        if connectivity == 6:
            joints = face_joints
        elif connectivity == 6 + 8:
            joints = face_joints + corner_joints
        elif connectivity == 6 + 12:
            joints = face_joints + edge_joints
        elif connectivity == 12 + 8:
            joints = edge_joints + corner_joints
        elif connectivity == 6 + 8 + 12:
            joints = face_joints + corner_joints + edge_joints
        else:
            raise Exception(
                "Invalid connectivity receieved. Must be 6, 14, 18, 20 or 26"
            )

        tails, heads = np.array([], dtype=int), np.array([], dtype=int)
        for T, H in joints:
            tails = np.concatenate((tails, T.flatten()))
            heads = np.concatenate((heads, H.flatten()))
        pairs = np.vstack([tails, heads]).T

        super().__init__(
            Np=points.shape[0], Nt=pairs.shape[0], name=name, project=project
        )

        self["pore.coords"] = points
        self["throat.conns"] = pairs
        self["pore.internal"] = True
        self["throat.internal"] = True
        self._label_surface_pores()
        topotools.label_faces(network=self)
        Ps = self["pore.surface"]
        self["throat.surface"] = np.all(Ps[self["throat.conns"]], axis=1)
        # Scale network to requested spacing
        self["pore.coords"] *= spacing

    def _label_surface_pores(self):
        r"""
        """
        hits = np.zeros_like(self.Ps, dtype=bool)
        dims = topotools.dimensionality(self)
        mn = np.amin(self["pore.coords"], axis=0)
        mx = np.amax(self["pore.coords"], axis=0)
        for ax in [0, 1, 2]:
            if dims[ax]:
                hits += self["pore.coords"][:, ax] <= mn[ax]
                hits += self["pore.coords"][:, ax] >= mx[ax]
        self["pore.surface"] = hits

    def add_boundary_pores(
        self, labels=["top", "bottom", "front", "back", "left", "right"], spacing=None
    ):
        r"""
        Add pores to the faces of the network for use as boundary pores.

        Pores are offset from the faces by 1/2 a lattice spacing such that
        they lie directly on the boundaries.

        Parameters
        ----------
        labels : string or list of strings
            The labels indicating the pores defining each face where boundary
            pores are to be added (e.g. 'left' or ['left', 'right'])

        spacing : scalar or array_like
            The spacing of the network (e.g. [1, 1, 1]).  This should be given
            since it can be quite difficult to infer from the network, for
            instance if boundary pores have already added to other faces.

        """
        if type(labels) == str:
            labels = [labels]
        x, y, z = self["pore.coords"].T
        if spacing is None:
            spacing = self._get_spacing()
        else:
            spacing = np.array(spacing)
            if spacing.size == 1:
                spacing = np.ones(3) * spacing
        Lcx, Lcy, Lcz = spacing

        offset = {}
        offset["front"] = offset["left"] = offset["bottom"] = [0, 0, 0]
        offset["back"] = [Lcx * self._shape[0], 0, 0]
        offset["right"] = [0, Lcy * self._shape[1], 0]
        offset["top"] = [0, 0, Lcz * self._shape[2]]

        scale = {}
        scale["front"] = scale["back"] = [0, 1, 1]
        scale["left"] = scale["right"] = [1, 0, 1]
        scale["bottom"] = scale["top"] = [1, 1, 0]

        for label in labels:
            try:
                Ps = self.pores(label)
                topotools.clone_pores(
                    network=self, pores=Ps, labels=label + "_boundary"
                )
                # Translate cloned pores
                ind = self.pores(label + "_boundary")
                coords = self["pore.coords"][ind]
                coords = coords * scale[label] + offset[label]
                self["pore.coords"][ind] = coords
            except KeyError:
                logger.warning(
                    "No pores labelled "
                    + label
                    + " were found, "
                    + "skipping boundary addition"
                )

    def _get_spacing(self):
        # Find Network spacing
        P12 = self["throat.conns"]
        C12 = self["pore.coords"][P12]
        mag = np.linalg.norm(np.diff(C12, axis=1), axis=2)
        unit_vec = np.around(np.squeeze(np.diff(C12, axis=1)) / mag, decimals=14)
        spacing = [0, 0, 0]
        dims = topotools.dimensionality(self)
        # Ensure vectors point in n-dims unique directions
        c = {tuple(row): 1 for row in unit_vec}
        if len(c.keys()) > sum(dims):
            raise Exception(
                "Spacing is undefined when throats point in "
                + "more directions than network has dimensions"
            )
        mag = sp.float64(mag.squeeze())
        for ax in [0, 1, 2]:
            if dims[ax]:
                inds = np.where(unit_vec[:, ax] == unit_vec[:, ax].max())[0]
                temp = np.unique(mag[inds])
                if not np.allclose(temp, temp[0]):
                    raise Exception("A unique value of spacing could not be found")
                else:
                    spacing[ax] = temp[0]
        return np.array(spacing)

    spacing = property(fget=_get_spacing)

    def _get_shape(self):
        L = np.ptp(self["pore.coords"], axis=0)
        mask = L.astype(bool)
        S = self.spacing
        shape = np.array([1, 1, 1], int)
        shape[mask] = L[mask] / S[mask] + 1
        return shape

    shape = property(fget=_get_shape)

    def to_array(self, values):
        r"""
        Converts the values to a rectangular array with the same shape as the
        network

        Parameters
        ----------
        values : array_like
            An Np-long array of values to convert to

        Notes
        -----
        This method can break on networks that have had boundaries added.  It
        will usually work IF the given values came only from 'internal'
        pores.

        """
        if np.shape(values)[0] > self.num_pores("internal"):
            raise Exception("The array shape does not match the network")
        Ps = np.array(self["pore.index"][self.pores("internal")], dtype=int)
        arr = np.ones(self._shape) * sp.nan
        ind = np.unravel_index(Ps, self._shape)
        arr[ind[0], ind[1], ind[2]] = values
        return arr

    def from_array(self, array, propname):
        r"""
        Apply data to the network based on a rectangular array filled with
        values.  Each array location corresponds to a pore in the network.

        Parameters
        ----------
        array : array_like
            The rectangular array containing the values to be added to the
            network. This array must be the same shape as the original network.

        propname : string
            The name of the pore property being added.

        """
        array = np.atleast_3d(array)
        if np.shape(array) != self._shape:
            raise Exception("The array shape does not match the network")
        temp = array.flatten()
        Ps = np.array(self["pore.index"][self.pores("internal")], dtype=int)
        propname = "pore." + propname.split(".")[-1]
        self[propname] = sp.nan
        self[propname][self.pores("internal")] = temp[Ps]



import openpnm.models as mods
from openpnm.geometry import GenericGeometry
from openpnm.utils import logging, GenericSettings
logger = logging.getLogger(__name__)


class ImportedSettings(GenericSettings):
    r"""

    Parameters
    ----------

    pore_diameter : str (default = 'pore.extended_diameter')
        Key into the extracted data array to use as pore diameter in other
        geometry calculations. The default is .  Use of 'pore.' is not
        required.
    throat_diameter : str (default = 'throat.equivalent_diameter')
        Key into the extracted data array to use as throat diameter in other
        geometry calculations. Use of 'throat.' is not required.
    pore_shape : string {'sphere' (default), 'cube'}
        Specifies which shape to assume when calculating dependent properties
        such as volume and surface area.
    throat_shape : string {'cylinder' (default), 'cuboid'}
        Specifies which shape to assume when calculating dependent properties
        such as volume and surface area.

    """
    pore_diameter = 'pore.extended_diameter'
    throat_diameter = 'throat.equivalent_diameter'
    pore_shape = 'sphere'
    throat_shape = 'cylinder'


class Imported(GenericGeometry):
    r"""
    This geometry class extracts all numerical properites from the received
    network object and moves them to itself.

    This class is intended for use with networks imported from network
    extraction codes, where the geometry properties are included on the
    network itself.

    Parameters
    ----------
    network : OpenPNM Network object
        The network with which this Geometry should be associated

    exclude : list of strings
        A list of which network properties should *not* be transferred to
        new geometry object.  'pore.coords' and 'throat.conns' are *always*
        excluded.  Note that labels are not transferred, only properties.

    project : OpenPNM Project object, optional
        Can be supplied in addition to ``network`` but is inferred from the
        network's project if not given.

    name : string
        The name of the object, which is also used as the label where this
        geometry is defined.

    Notes
    -----
    An error occurs when adding other geometries to a network that has
    geometrical properties such as 'pore.diameter'.  This can occur when
    adding boundary pores or in more elaborate scenarios such as stitching
    networks together.  The issue arises because OpenPNM prevents a property,
    such as 'pore.volume', from existing on both the network and also a
    geometry.  Thus it is necessary to move the extracted network properties
    to this ``Imported`` class, then create new geometry objects for any
    added pores as needed.

    """

    def __init__(self, network, exclude=[], settings={}, **kwargs):
        self.settings._update_settings_and_docs(ImportedSettings())
        self.settings.update(settings)
        super().__init__(network=network, **kwargs)
        # Transfer all geometrical properties off of network
        exclude.extend(['pore.coords', 'throat.conns'])
        for item in network.props():
            if item not in exclude:
                self[item] = network.pop(item)

        # If the following 'essential' props are not already defined, then
        # they should be added using the specified values or models

        if 'pore.diameter' not in self.keys():
            pdia = 'pore.'+self.settings['pore_diameter'].split('pore.')[-1]
            try:
                self['pore.diameter'] = self[pdia]
            except KeyError:
                logger.error(pdia + " not found, can't assign 'pore.diameter'")

        if 'pore.volume' not in self.keys():
            pore_shape = self.settings['pore_shape']
            m = getattr(mods.geometry.pore_volume, pore_shape)
            self.add_model(propname='pore.volume',
                           model=m, pore_diameter='pore.diameter')

        if 'pore.area' not in self.keys():
            pore_shape = self.settings['pore_shape']
            m = getattr(mods.geometry.pore_area, pore_shape)
            self.add_model(propname='pore.area',
                           model=m)

        if 'throat.diameter' not in self.keys():
            tdia = 'throat.'+self.settings['throat_diameter'].split('throat.')[-1]
            try:
                self['throat.diameter'] = self[tdia]
            except KeyError:
                logger.error(tdia + " not found, can't assign 'throat.diameter'")

        if 'throat.endpoints' not in self.keys():
            self.add_model(propname='throat.endpoints',
                           model=mods.geometry.throat_endpoints.spherical_pores,
                           pore_diameter='pore.diameter',
                           throat_diameter='throat.diameter')

        if 'throat.length' not in self.keys():
            self.add_model(propname='throat.length',
                           model=mods.geometry.throat_length.piecewise,
                           throat_endpoints='throat.endpoints')

        if 'throat.volume' not in self.keys():
            shape = self.settings['throat_shape']
            m = getattr(mods.geometry.throat_volume, shape)
            self.add_model(propname='throat.volume',
                           model=m,
                           throat_length='throat.length',
                           throat_diameter='throat.diameter')

        if 'throat.conduit_lengths' not in self.keys():
            self.add_model(propname='throat.conduit_lengths',
                           model=mods.geometry.throat_length.conduit_lengths,
                           throat_endpoints='throat.endpoints',
                           throat_length='throat.length')



import numpy as _np


def mix_and_match(target, prop, phases, occupancy):
    r"""
    Return the given property by looking it up from a list of given phases
    based on occupancy.

    Parameters
    ----------
    target : OpenPNM Object
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.

    prop : string
        The dictionary key to the array containing the pore/throat property to
        be used in the calculation.

    phases : list
        List of OpenPNM phase objects over which the given `prop` is to be
        averaged out.

    occupancy : string
        The dictionary key to the array containing the occupancy associated
        with each of the given ``phases``.

    Returns
    -------
    weighted_average : ndarray
        Weighted average of the given `prop` averaged over `phases`.

    """
    # Hack for ModelsMixin to not complain (cyclic dep)
    prop = prop.strip("_")

    values = _np.zeros_like(phases[0][prop])

    for phase in phases:
        mask = phase[occupancy]
        values[mask] = phase[prop][mask]

    return values



import os
import json
import pickle
import numpy as np
import scipy as sp
from pathlib import Path
from openpnm.utils import logging
from openpnm.io import GenericIO
import openpnm.models.geometry as gmods
from openpnm.network import GenericNetwork
logger = logging.getLogger(__name__)


class JSONGraphFormat(GenericIO):
    r"""
    Class for reading and writing OpenPNM networks to JSON Graph Format (JGF).

    Notes
    -----
    The JGF standard must contain data formatted according to
    http://jsongraphformat.info and enforced by JSON schema validation.

    Users must transfer any phase data to the network manually if they wish to
    keep it.
    """

    @classmethod
    def __validate_json__(self, json_file):
        import jsonschema
        # Validate name of schema file
        relative_path = '../../utils/jgf_schema.pkl'
        schema_file = Path(os.path.realpath(__file__), relative_path)
        schema_file = self._parse_filename(filename=schema_file, ext='pkl')

        # Load schema from pickle file
        with open(schema_file, 'rb') as file:
            jgf_schema = pickle.load(file)

        # Validate JSON agains schema
        try:
            jsonschema.validate(json_file, jgf_schema)
            return True
        except jsonschema.exceptions.ValidationError:
            return False

    @classmethod
    def save(self, network, filename=''):
        r"""
        Write the wetwork to disk as a JGF file.

        Parameters
        ----------
        network : OpenPNM Network Object

        filename : string
            Desired file name, defaults to network name if not given
        """

        # Ensure output file is valid
        filename = self._parse_filename(filename=filename, ext='json')

        # Ensure network contains the required properties
        try:
            required_props = {'pore.diameter', 'pore.coords', 'throat.length',
                              'throat.conns', 'throat.diameter'}
            assert required_props.issubset(network.props())
        except AssertionError:
            raise Exception('Error - network is missing one of: '
                            + str(required_props))

        # Create 'metadata' JSON object
        graph_metadata_obj = {'number_of_nodes': network.Np,
                              'number_of_links': network.Nt}

        # Create 'nodes' JSON object
        nodes_obj = [
            {
                'id': str(ps),
                'metadata': {
                    'node_squared_radius': int(network['pore.diameter'][ps]/2)**2,
                    'node_coordinates': {
                        'x': int(network['pore.coords'][ps, 0]),
                        'y': int(network['pore.coords'][ps, 1]),
                        'z': int(network['pore.coords'][ps, 2])
                    }
                }
            } for ps in network.Ps]

        # Create 'edges' JSON object
        edges_obj = [
            {
                'id': str(ts),
                'source': str(network['throat.conns'][ts, 0]),
                'target': str(network['throat.conns'][ts, 1]),
                'metadata': {
                    'link_length': float(network['throat.length'][ts]),
                    'link_squared_radius': float(network['throat.diameter'][ts]/2)**2
                }
            } for ts in network.Ts]

        # Build 'graph' JSON object from 'metadata', 'nodes' and 'edges'
        graph_obj = {'metadata': graph_metadata_obj,
                     'nodes': nodes_obj,
                     'edges': edges_obj}

        # Build full JSON object
        json_obj = {'graph': graph_obj}

        # Load and validate input JSON
        with open(filename, 'w') as file:
            json.dump(json_obj, file, indent=2)

    @classmethod
    def load(self, filename, project=None):
        r"""
        Loads the JGF file onto the given project.

        Parameters
        ----------
        filename : string
            The name of the file containing the data to import.  The formatting
            of this file is outlined below.

        project : OpenPNM Project object
            A GenericNetwork is created and added to the specified Project.
            If no Project object is supplied then one will be created and
            returned.

        Returns
        -------
        If no project object is supplied then one will be created and returned.
        """

        # Ensure input file is valid
        filename = self._parse_filename(filename=filename, ext='json')

        # Load and validate input JSON
        with open(filename, 'r') as file:
            json_file = json.load(file)
            if not self.__validate_json__(json_file):
                raise Exception('FIle is not in the JSON Graph Format')

        # Extract graph metadata from JSON
        number_of_nodes = json_file['graph']['metadata']['number_of_nodes']
        number_of_links = json_file['graph']['metadata']['number_of_links']

        # Extract node properties from JSON
        nodes = sorted(json_file['graph']['nodes'], key=lambda node: int(node['id']))
        x = np.array([node['metadata']['node_coordinates']['x'] for node in nodes])
        y = np.array([node['metadata']['node_coordinates']['y'] for node in nodes])
        z = np.array([node['metadata']['node_coordinates']['z'] for node in nodes])

        # Extract link properties from JSON
        edges = sorted(json_file['graph']['edges'], key=lambda edge: int(edge['id']))
        source = np.array([int(edge['source']) for edge in edges])
        target = np.array([int(edge['target']) for edge in edges])
        link_length = np.array([edge['metadata']['link_length'] for edge in edges])
        link_squared_radius = np.array(
            [edge['metadata']['link_squared_radius'] for edge in edges])

        # Generate network object
        network = GenericNetwork(Np=number_of_nodes, Nt=number_of_links)

        # Define primitive throat properties
        network['throat.length'] = link_length
        network['throat.conns'] = np.column_stack([source, target])
        network['throat.diameter'] = 2.0 * np.sqrt(link_squared_radius)

        # Define derived throat properties
        network['throat.area'] = gmods.throat_area.cylinder(network)
        network['throat.volume'] = gmods.throat_volume.cylinder(network)
        network['throat.perimeter'] = gmods.throat_perimeter.cylinder(network)
        network['throat.surface_area'] = gmods.throat_surface_area.cylinder(network)

        # Define primitive pore properties
        network['pore.index'] = np.arange(number_of_nodes)
        network['pore.coords'] = np.column_stack([x, y, z])
        network['pore.diameter'] = np.zeros(number_of_nodes)

        # Define derived pore properties
        network['pore.area'] = gmods.pore_area.sphere(network)
        network['pore.volume'] = gmods.pore_volume.sphere(network)

        return network.project



r"""

**openpnm.physics**

----

This module contains the GenericPhysics class, along with a few preconfigured
classes that include common pore-scale physical models

----

**The GenericPhysics Class**

----

**Library of Preconfigured Physics Classes**

+-------------+---------------------------------------------------------------+
| Class       | Comments                                                      |
+=============+===============================================================+
| Standard    |                                                               |
+-------------+---------------------------------------------------------------+

"""

from .GenericPhysics import GenericPhysics
from .Standard import Standard
from .Standard_2D import Standard_2D
from .Classic import Classic



r"""

.. autofunction:: openpnm.models.geometry.throat_perimeter.cylinder
.. autofunction:: openpnm.models.geometry.throat_perimeter.cuboid

"""
import scipy as _sp


def cylinder(target, throat_diameter='throat.diameter'):
    r"""
    Calcuate the throat perimeter assuming a circular cross-section

    Parameters
    ----------
    target : OpenPNM Object
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.

    throat_diameter : string
        The dictionary key of the array containing the throat diameter values

    Returns
    -------
    value : NumPy ndarray
        Array containing throat perimeter values.

    """
    return target[throat_diameter]*_sp.pi


def cuboid(target, throat_diameter='throat.diameter'):
    r"""
    Calcuate the throat perimeter assuming a square cross-section

    Parameters
    ----------
    target : OpenPNM Object
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.

    throat_diameter : string
        The dictionary key of the array containing the throat diameter values.

    Returns
    -------
    value : NumPy ndarray
        Array containing throat perimeter values.

    """
    return target[throat_diameter]*4


def rectangle(target, throat_diameter='throat.diameter'):
    r"""
    Calcuate the throat perimeter assuming a rectangular cross-section (2D)

    Parameters
    ----------
    target : OpenPNM Object
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.

    throat_diameter : string
        The dictionary key of the array containing the throat diameter values

    Returns
    -------
    value : NumPy ndarray
        Array containing throat perimeter values.

    """
    return 1.0



# from collections import ChainMap  # Might use eventually
from openpnm.phases.mixtures import GenericMixture
import openpnm.models as mods
from openpnm.utils import logging, Docorator
docstr = Docorator()
logger = logging.getLogger(__name__)


@docstr.dedent
class IdealGas(GenericMixture):
    r"""
    Creates Mixture object that represents a ideal gas system
    consisting of a given list of OpenPNM Phase objects as components.

    Parameters
    ----------
    %(GenericMixture.parameters)s

    """

    def __init__(self, settings={}, **kwargs):
        super().__init__(settings={'prefix': 'mix'}, **kwargs)
        self.settings.update(settings)

        self.add_model(propname='pore.molar_density',
                       model=mods.phases.molar_density.ideal_gas,
                       regen_mode='deferred')



import numpy as np
import scipy as sp
import openpnm as op
from pathlib import Path
from openpnm.utils import flat_list, logging

logger = logging.getLogger(__name__)


class GenericIO:
    @classmethod
    def save(cls):
        raise NotImplementedError(
            "The 'save' method for this class " + "does not exist yet"
        )

    @classmethod
    def load(cls):
        raise NotImplementedError(
            "The 'load' method for this class " + "does not exist yet"
        )

    @classmethod
    def _convert_data(cls, project):
        # Convert arrays of 1's and/or 0's to booleans
        for obj in project:
            for item in obj.keys():
                ra = obj[item]
                if (ra.dtype == int) and (ra.max() <= 1) and (ra.min() >= 0):
                    obj.update({item: ra.astype(bool)})
        return project

    @classmethod
    def _update_network(cls, network, net):
        # Infer Np and Nt from length of given prop arrays in file
        for el in ["pore", "throat"]:
            N = [np.shape(net[i])[0] for i in net.keys() if i.startswith(el)]
            if N:
                N = np.array(N)
                if np.all(N == N[0]):
                    if (network._count(el) == N[0]) or (network._count(el) == 0):
                        network.update({el + ".all": np.ones((N[0],), dtype=bool)})
                        net.pop(el + ".all", None)
                    else:
                        raise Exception(
                            "Length of "
                            + el
                            + " data in file"
                            + " does not match network"
                        )
                else:
                    raise Exception(el + " data in file have inconsistent" + " lengths")
        # Add data on dummy net to actual network
        for item in net.keys():
            # Try to infer array types and change if necessary
            # Chcek for booleans disguised and 1's and 0's
            if (net[item].dtype is int) and (net[item].max() == 1):
                net[item] = net[item].astype(bool)
            # Write data to network object
            network.update({item: net[item]})

        network._gen_ids()

        return network

    @classmethod
    def _parse_filename(cls, filename, ext=""):
        p = Path(filename)
        p = p.resolve()
        if ext == "":
            ext = p.suffix
        # If extension not part of filename
        ext = "." + (ext.strip("."))
        if p.suffix != ext:
            p = p.with_suffix(ext)
        return p

    @classmethod
    def _parse_args(cls, network, phases):
        # Convert network to a list, even if empty
        if network is None:
            network = []
        else:
            network = flat_list(network)
        # Ensure phases is a list, even if empty
        phases = flat_list(phases)
        # Get handle to project object
        if len(network) == 0:
            if len(phases) == 0:
                raise Exception("Must specify one of network or phase")
            else:
                project = phases[0].project
        else:
            project = network[0].project
        return (project, network, phases)

    @classmethod
    def is_transient(cls, phases):
        # Check if any of the phases has time series
        transient = False
        if type(phases) == str:
            transient = True in ["@" in k for k in phases.keys()]
        elif type(phases) == list:
            for phase in phases:
                transient = True in ["@" in k for k in phase.keys()]
                if transient:
                    break
        return transient



from openpnm.algorithms import TransientReactiveTransport, FickianDiffusion
from openpnm.utils import logging, Docorator
logger = logging.getLogger(__name__)
docstr = Docorator()


class TransientFickianDiffusion(TransientReactiveTransport, FickianDiffusion):
    r"""
    A class to simulate transient diffusion with reactions

    """

    def __init__(self, settings={}, **kwargs):
        super().__init__(**kwargs)
        self.settings.update(settings)

    @docstr.dedent
    def setup(self, phase=None, quantity='', conductance='',
              t_initial=None, t_final=None, t_step=None, t_output=None,
              t_tolerance=None, t_scheme='', **kwargs):
        r"""

        Parameters
        ----------
        %(FickianDiffusionSettings.parameters)s

        Notes
        -----
        Any additional arguments are added to the ``settings`` dictionary of
        the object.

        """
        if phase:
            self.settings['phase'] = phase.name
        if quantity:
            self.settings['quantity'] = quantity
        if conductance:
            self.settings['conductance'] = conductance
        if t_initial:
            self.settings['t_initial'] = t_initial
        if t_final:
            self.settings['t_final'] = t_final
        if t_step:
            self.settings['t_step'] = t_step
        if t_output is not None:
            self.settings['t_output'] = t_output
        if t_tolerance:
            self.settings['t_tolerance'] = t_tolerance
        if t_scheme:
            self.settings['t_scheme'] = t_scheme
        self.settings.update(kwargs)



from openpnm.core import Base
from openpnm.utils import logging
import numpy as np
logger = logging.getLogger(__name__)


class GenericMetric(Base):
    r"""

    """

    def __init__(self, network=None, project=None, settings={}, **kwargs):
        self.settings['prefix'] = 'metric'
        super().__init__(project=project, **kwargs)
        project = self.project
        self['pore.all'] = np.ones((project.network.Np, ), dtype=bool)
        self['throat.all'] = np.ones((project.network.Nt, ), dtype=bool)



from openpnm.phases import GenericPhase as GenericPhase
from openpnm.utils import logging
logger = logging.getLogger(__name__)


class GenericSpecies(GenericPhase):
    r"""
    Creates Phase object that represents a single species in a multicomponent
    mixture system.

    Parameters
    ----------
    network : OpenPNM Network object
        The network to which this phase object will be attached.

    project : OpenPNM Project object, optional
        The Project with which this phase should be associted.  If a
        ``network`` is given then this is ignored and the Network's project
        is used.  If a ``network`` is not given then this is mandatory.

    name : string, optional
        The name of the phase.  This is useful to keep track of the objects
        throughout the simulation.  The name must be unique to the project.
        If no name is given, one is generated.

    """
    @property
    def mixture(self):
        for item in self.project:
            if hasattr(item, 'components'):
                for comp in item.components.values():
                    if self is comp:
                        return item



r"""

.. autofunction:: openpnm.models.physics.electrical_conductance.series_resistors

"""
import numpy as _np
import scipy as _sp


def series_resistors(target,
                     pore_area='pore.area',
                     throat_area='throat.area',
                     pore_conductivity='pore.electrical_conductivity',
                     throat_conductivity='throat.electrical_conductivity',
                     conduit_lengths='throat.conduit_lengths',
                     conduit_shape_factors='throat.poisson_shape_factors'):
    r"""
    Calculate the electrical conductance of conduits in network, where a
    conduit is ( 1/2 pore - full throat - 1/2 pore ). See the notes section.

    Parameters
    ----------
    target : OpenPNM Object
        The object which this model is associated with. This controls the
        length of the calculated array, and also provides access to other
        necessary properties.

    pore_area : string
        Dictionary key of the pore area values

    throat_area : string
        Dictionary key of the throat area values

    pore_conductivity : string
        Dictionary key of the pore thermal conductivity values

    throat_conductivity : string
        Dictionary key of the throat thermal conductivity values

    conduit_lengths : string
        Dictionary key of the conduit length values

    conduit_shape_factors : string
        Dictionary key of the conduit DIFFUSION shape factor values

    Returns
    -------
    g : ndarray
        Array containing electrical conductance values for conduits in the
        geometry attached to the given physics object.

    Notes
    -----
    (1) This function requires that all the necessary phase properties already
    be calculated.

    (2) This function calculates the specified property for the *entire*
    network then extracts the values for the appropriate throats at the end.

    (3) This function assumes cylindrical throats with constant cross-section
    area. Corrections for different shapes and variable cross-section area can
    be imposed by passing the proper conduit_shape_factors argument.

    (4) shape_factor depends on the physics of the problem, i.e. diffusion-like
    processes and fluid flow need different shape factors.

    """
    network = target.project.network
    throats = network.map_throats(throats=target.Ts, origin=target)
    phase = target.project.find_phase(target)
    cn = network['throat.conns'][throats]
    # Getting equivalent areas
    A1 = network[pore_area][cn[:, 0]]
    At = network[throat_area][throats]
    A2 = network[pore_area][cn[:, 1]]
    # Getting conduit lengths
    L1 = network[conduit_lengths + '.pore1'][throats]
    Lt = network[conduit_lengths + '.throat'][throats]
    L2 = network[conduit_lengths + '.pore2'][throats]
    # Preallocating g
    g1, g2, gt = _np.zeros((3, len(Lt)))
    # Setting g to inf when Li = 0 (ex. boundary pores)
    # INFO: This is needed since area could also be zero, which confuses NumPy
    m1, m2, mt = [Li != 0 for Li in [L1, L2, Lt]]
    g1[~m1] = g2[~m2] = gt[~mt] = _sp.inf
    # Getting shape factors
    try:
        SF1 = phase[conduit_shape_factors+'.pore1'][throats]
        SFt = phase[conduit_shape_factors+'.throat'][throats]
        SF2 = phase[conduit_shape_factors+'.pore2'][throats]
    except KeyError:
        SF1 = SF2 = SFt = 1.0
    # Interpolate pore phase property values to throats
    try:
        Dt = phase[throat_conductivity][throats]
    except KeyError:
        Dt = phase.interpolate_data(propname=pore_conductivity)[throats]
    try:
        D1 = phase[pore_conductivity][cn[:, 0]]
        D2 = phase[pore_conductivity][cn[:, 1]]
    except KeyError:
        D1 = phase.interpolate_data(propname=throat_conductivity)[cn[:, 0]]
        D2 = phase.interpolate_data(propname=throat_conductivity)[cn[:, 1]]
    # Find g for half of pore 1, throat, and half of pore 2
    g1[m1] = (D1*A1)[m1] / L1[m1]
    g2[m2] = (D2*A2)[m2] / L2[m2]
    gt[mt] = (Dt*At)[mt] / Lt[mt]
    # Apply shape factors and calculate the final conductance
    return (1/gt/SFt + 1/g1/SF1 + 1/g2/SF2)**(-1)
